import React, { useMemo } from "react";
import { useNavigation } from "@react-navigation/native";
import { TouchableOpacity } from "react-native";
import useColors from "../scripts/useColors";
import Icon from "./Icon";
import { BlurView } from "expo-blur";
import usePaddedSafeAreaInsets from "../scripts/usePaddedSafeAreaInsets";

/**
 * @param {object} props
 * @param {import("react-native").StyleProp<import("react-native").ViewStyle>} [props.style] Styles
 *      to apply to the containing <View>.
 * @param {Number} [props.blurIntensity] A number in the range [0,100] that controls the intensity
 *      of the blur. A lower number will have very little blur (with 0 being no blur) and higher
 *      numbers will have more blur (with 100 being completely opaque). Default is 65.
 * @param {(import("react-native").GestureResponderEvent) => void} [props.onPress] The function that
 *      gets called when the button is pressed. If omitted, the button will perform
 *      `navigation.goBack()`.
 */
const BackBtn = ({style, blurIntensity, onPress, ...props}) => {
    const navigation = useNavigation();
    const colors = useColors();
    const insets = usePaddedSafeAreaInsets();
    blurIntensity = typeof blurIntensity == "number" && 0 <= blurIntensity && blurIntensity <= 100 ? blurIntensity : 65;
    const onPressHandler = useMemo(() =>
        typeof onPress == "function" ? onPress : () => void navigation.goBack()
    , [onPress]);

    return (
        <TouchableOpacity
            style={[
                {
                    position: "absolute",
                    left: insets.left,
                    top: insets.top,
                    zIndex: 1,
                    borderRadius: "50%",
                    overflow: "hidden"
                },
                style
            ]}
            onPress={onPressHandler}
        >
            <BlurView
                tint={colors.COLOR_SCHEME}
                intensity={blurIntensity}
                style={{
                    padding: 10,
                    aspectRatio: 1/1,
                    justifyContent: "center",
                    alignItems: "center"
                }}
            >
                <Icon height={24}>chevron-left</Icon>
            </BlurView>
        </TouchableOpacity>
    )
};

export default BackBtn;