import React, { ReactNode, useState } from "react";
import { Pressable, StyleProp, TextProps, TextStyle, ViewProps, ViewStyle } from "react-native";
import useColors from "../scripts/useColors";
import Icon, { IconProps } from "./Icon";
import StyledText from "./StyledText";

interface ClickableTextProps extends TextProps, ViewProps {
    style?: StyleProp<TextStyle>;
    /**
     * Whether to show this text as disabled. If it is disabled, the text will no longer be
     * clickable and it will show up in a gray text color instead.
     */
    disabled?: boolean | undefined;
    children?: ReactNode | String;
    /**
     * An optional icon to display before the text. This should be the icon's name.
     */
    beforeIcon?: String | undefined;
    /**
     * An option icon to display after the text. This should be the icon's name.
     */
    afterIcon?: String | undefined;
    /**
     * Styles to apply to both icons (`beforeIcon` and `afterIcon`).
     */
    iconStyle?: IconProps["style"];
    /**
     * Styles to apply to `beforeIcon`.
     */
    beforeIconStyle?: StyleProp<ViewStyle>;
    /**
     * Styles to apply to `afterIcon`.
     */
    afterIconStyle?: StyleProp<ViewStyle>;
    /**
     * Props to apply to both icon (`beforeIcon` and `afterIcon`).
     */
    iconProps?: IconProps | undefined;
    /**
     * Props to apply to `beforeIcon`.
     */
    beforeIconProps?: IconProps | undefined;
    /**
     * Props to apply to `afterIcon`.
     */
    afterIconProps?: IconProps | undefined;
};

export const ClickableText = (
    {
        children,
        disabled,
        style,
        beforeIcon,
        afterIcon,
        iconStyle,
        beforeIconStyle,
        afterIconStyle,
        iconProps,
        beforeIconProps,
        afterIconProps,
        onPress,
        onPressIn,
        onPressOut,
        ...props
    }: ClickableTextProps
) => {
    let [isBeingPressed, setIsBeingPressed] = useState(false);
    const colors = useColors();

    const childrenArray =
        typeof children == "string" ? [children] :
        Array.from(children as Iterable<ReactNode | String>) || [];

    const font_styles: StyleProp<TextStyle> = {
        color: disabled ? colors.SECONDARY_TEXT : colors.FLAT2,
        fontWeight: "500",
        textDecorationLine: isBeingPressed || disabled ? "none" : "underline",
        textDecorationStyle: "solid",
    };
    const all_styles: StyleProp<ViewStyle> = style || {};
    for (const style_name of [
        "color",
        "fontFamily",
        "fontSize",
        "fontStyle",
        "fontVariant",
        "fontWeight",
        "includeFontPadding",
        "letterSpacing",
        "lineHeight",
        "textAlign",
        "textAlignVertical",
        "textDecorationColor",
        "textDecorationLine",
        "textDecorationStyle",
        "textShadowColor",
        "textShadowOffset",
        "textShadowRadius",
        "textTransform",
        "writingDirection",
    ]) {
        if (style_name in all_styles) {
            font_styles[style_name] = all_styles[style_name];
        }
    }

    // Replace all plain strings in the children with <Text>s.
    for (let i = 0; i < childrenArray.length; i++) {
        if (typeof childrenArray[i] == "string") {
            childrenArray[i] = (
                <StyledText key={i} style={font_styles}>
                    {childrenArray[i]}
                </StyledText>
            );
        }
    }

    // Add in the beforeIcon and afterIcon if one is provided.
    if (typeof beforeIcon == "string") {
        childrenArray.unshift(
            <Icon
                key={-1}
                color={font_styles.color}
                height={"fontSize" in font_styles ? font_styles.fontSize : 16}
                style={[iconStyle, beforeIconStyle]}
                {...iconProps}
                {...beforeIconProps}
            >{beforeIcon}</Icon>
        );
    }
    if (typeof afterIcon == "string") {
        childrenArray.push(
            <Icon
                key={childrenArray.length}
                color={font_styles.color}
                height={"fontSize" in font_styles ? font_styles.fontSize : 16}
                style={[iconStyle, afterIconStyle]}
                {...iconProps}
                {...afterIconProps}
            >{afterIcon}</Icon>
        );
    }

    return (
        <Pressable
            {...props}
            onPressIn={(e) => {
                setIsBeingPressed(true);
                if (typeof onPressIn == "function") {
                    onPressIn(e);
                }
            }}
            onPressOut={(e) => {
                setIsBeingPressed(false);
                if (typeof onPressOut == "function") {
                    onPressOut(e);
                }
            }}
            onPress={(e) => {
                if (typeof onPress == "function") {
                    onPress(e);
                }
            }}
            {
                // If the button is disabled, overwrite the onPressIn, onPressOut, and onPress
                // events to just null so that the button doesn't have any interactions when the
                // user presses it.
                ...(disabled
                    ? {
                          onPressIn: null,
                          onPressOut: null,
                          onPress: null,
                      }
                    : {}
                )
            }
            style={[
                {
                    display: "flex",
                    flexDirection: "row",
                    alignItems: "center",
                    justifyContent: "center",
                },
                style
            ]}
        >
            {childrenArray}
        </Pressable>
    );
};