import React, { useState } from "react";
import { Image, View } from "react-native";
import Color from "color";
import StyledText from "./StyledText";
import { interestCategories } from "../scripts/hardcoded_data";
import useColors from "../scripts/useColors";
import { ClickableText } from "./ClickableText";

/**
 * The maximum number of interests to show by default. The user can opt to click Show more to see
 * all the interests.
 * @constant
 * @type {Number}
 * @default 7
 */
const MAX_INTERESTS = 7;

/**
 * A list of interests. Each interest is show in a bubble with the name of the interest and an icon.
 * If there are too many interests to show at once, a Show More button will allow the user to
 * display all the interests.
 * @param {object} props
 * @param {boolean} props.similar Whether this list of interests corresponds to similar interests
 *      or not. Similar interests are displayed in `colors.FLAT1` while non-similar interests are
 *      displayed in `colors.FLAT2`.
 * @param {Iterable<String>} props.interests The list of interests to display.
 */
const InterestList = ({similar, interests, ...props}) => {
    interests = Array.from(interests).map(name => interestCategories.find((interest) => interest.name == name));
    const [isExpanded, setIsExpanded] = useState(false);

    const moreThanMax = interests.length > MAX_INTERESTS;
    const numberShowing = isExpanded ? interests.length : Math.min(MAX_INTERESTS, interests.length);
    const bubbles = new Array(numberShowing);

    for (let i = 0; i < numberShowing; i++) {
        bubbles[i] = (
            <InterestBubble
                key={interests[i].id}
                isSimilarInterest={similar}
                interest={interests[i]}
            />
        );
    }

    return (
        <View style={{display: "flex", flexDirection: "row", flexWrap: "wrap", alignItems: "baseline"}}>
            {bubbles}
            {
                moreThanMax ?
                <ClickableText
                    onPress={e => setIsExpanded(!isExpanded)}
                    afterIcon={isExpanded ? "chevron-up" : "chevron-down"}
                    iconProps={{
                        height: 14
                    }}
                >{isExpanded ? "Show Less " : "Show More "}</ClickableText> :
                null
            }
        </View>
    );
};

/**
 * @param {object} props
 * @param {boolean} props.isSimilarInterest
 * @param {import("../scripts/hardcoded_data").Interest} props.interest
 */
const InterestBubble = ({isSimilarInterest, interest, ...props}) => {
    const colors = useColors();
    const backgroundColor = isSimilarInterest ? Color(colors.FLAT1).mix(Color(colors.BACKGROUND), 0.8).hex() : Color(colors.FLAT2).mix(Color(colors.BACKGROUND), 0.8).hex();
    const interestName = interest.name[0].toUpperCase() + interest.name.substring(1);

    return (
        <View style={{
            backgroundColor,
            borderRadius: "50%",
            alignItems: "center",
            paddingHorizontal: 20,
            paddingVertical: 10,
            marginTop: 6,
            marginRight: 8,
            display: "flex",
            flexDirection: "row"
        }}>
            <View style={{ height: 16 * 1.2, aspectRatio: 1/1 }}>
                <Image
                    style={{
                        position: "absolute",
                        height: "100%",
                        width: "100%"
                    }}
                    source={interest.imageUrl}
                />
            </View>
            <StyledText style={{fontSize: 16, fontWeight: 500}}> {interestName}</StyledText>
        </View>
    );
};

export default InterestList;