import { LinearGradient } from "expo-linear-gradient";
import React from "react";
import { Pressable, PressableProps, View, ViewStyle } from "react-native";
import useColors from "../scripts/useColors";
import StyledText from "./StyledText";
import MaskedView from '@react-native-masked-view/masked-view';

interface RowButtonProps extends PressableProps {
    /**
     * Whether this button should show up as a "main" action button or not. A "main" action button
     * will have its background be a gradient whereas a non-main button will only have a gradient
     * outline.
     */
    mainAction?: boolean | undefined;
    /**
     * Whether to display the button as a destructive button. The gradient colors will change to
     * indicate that the button is destructive.
     */
    destructive?: boolean | undefined;
    /**
     * Either children, a string of text, or a render prop that receives a boolean reflecting
     * whether the component is currently pressed.
     */
    children?: PressableProps["children"] | string;
}

/**
 * A pressable button that is made to take up an entire row in a list of buttons.
 */
const RowButton = ({mainAction, destructive, children, style, ...props}: RowButtonProps) => {
    const colors = useColors();
    const gradientColors = destructive ? [colors.FLAT3, colors.FLAT4] : [colors.FLAT1, colors.FLAT2];

    if (mainAction) {
        // If the children prop is just a single string, convert it into a <StyledText> with
        // appropriate styling.
        if (typeof children == "string") {
            children = (
                <StyledText style={{
                    color: colors.INVERTED_TEXT,
                    fontWeight: "700",
                    fontSize: 20
                }}>{children}</StyledText>
            );
        }

        return (
            <Pressable
                style={[
                    {
                        borderRadius: 999,
                        overflow: "hidden",
                        justifyContent: "center",
                        alignItems: "center",
                        flexGrow: 1,
                        display: "flex",
                        flexDirection: "row",
                        paddingVertical: 10,
                        paddingHorizontal: 30,
                    },
                    style as ViewStyle
                ]}
                {...props}
            >
                <LinearGradient
                    style={{
                        position: "absolute",
                        left: 0,
                        top: 0,
                        right: 0,
                        bottom: 0
                    }}
                    colors={gradientColors}
                    start={[1.3, .5]}
                    end={[0, .5]}
                />
                {children}
            </Pressable>
        );
    } else {
        if (typeof children == "string") {
            // If the children prop is just a single string, we want to render this text as part of
            // the overall gradient, so we need to incorporate it into the <MaskedView>.
            children = (
                <StyledText style={{
                    fontWeight: "700",
                    fontSize: 20
                }}>{children}</StyledText>
            );

            return (
                <Pressable
                    style={[
                        {
                            borderRadius: 999,
                            overflow: "hidden",
                            justifyContent: "center",
                            alignItems: "center",
                            flexGrow: 1,
                            display: "flex",
                            flexDirection: "row",
                            paddingVertical: 10,
                            paddingHorizontal: 30,
                            backgroundColor: colors.BACKGROUND
                        },
                        style as ViewStyle
                    ]}
                    {...props}
                >
                    <MaskedView
                        style={{
                            position: "absolute",
                            left: 0,
                            top: 0,
                            right: 0,
                            bottom: 0,
                            justifyContent: "center",
                            alignItems: "center"
                        }}
                        maskElement={
                            <View
                                style={{
                                    height: "100%",
                                    width: "100%",
                                    justifyContent: "center",
                                    borderWidth: 2.5,
                                    borderColor: "black",
                                    alignItems: "center",
                                    borderRadius: 999
                                }}
                            >
                                {children}
                            </View>
                        }
                    >
                        <LinearGradient
                            style={{
                                height: "100%",
                                width: "100%",
                            }}
                            colors={gradientColors}
                            start={[1.3, .5]}
                            end={[0, .5]}
                        />
                    </MaskedView>
                    <View
                        style={{
                            opacity: 0
                        }}
                    >
                        {children}
                    </View>
                </Pressable>
            )
        } else {
            // Otherwise, we can make the mask view only for the edges of the button.
            return (
                <Pressable
                    style={[
                        {
                            borderRadius: 999,
                            overflow: "hidden",
                            justifyContent: "center",
                            alignItems: "center",
                            flexGrow: 1,
                            display: "flex",
                            flexDirection: "row",
                            paddingVertical: 10,
                            paddingHorizontal: 30,
                            backgroundColor: colors.BACKGROUND
                        },
                        style as ViewStyle
                    ]}
                    {...props}
                >
                    <MaskedView
                        style={{
                            position: "absolute",
                            left: 0,
                            top: 0,
                            right: 0,
                            bottom: 0,
                            justifyContent: "center",
                            alignItems: "center"
                        }}
                        maskElement={
                            <View
                                style={{
                                    height: "100%",
                                    width: "100%",
                                    justifyContent: "center",
                                    borderWidth: 2.5,
                                    borderColor: "black",
                                    alignItems: "center",
                                    borderRadius: 999
                                }}
                            />
                        }
                    >
                        <LinearGradient
                            style={{
                                height: "100%",
                                width: "100%",
                            }}
                            colors={gradientColors}
                            start={[1.3, .5]}
                            end={[0, .5]}
                        />
                    </MaskedView>
                    {children}
                </Pressable>
            );
        }
    }
};

export default RowButton;