import React, { forwardRef, useCallback, useRef, useState } from "react";
import { NativeSyntheticEvent, TextInputFocusEventData, TouchableOpacity, ViewStyle } from "react-native";
import { View, TextInput, TextInputProps, StyleProp } from "react-native";
import { style_values } from "../scripts/styles";
import useColors from "../scripts/useColors";
import Icon from "./Icon";
import StyledText from "./StyledText";

interface SearchBarProps extends TextInputProps {
    /**
     * The string that will be rendered before text input has been entered. Default is `"Search"`.
     */
    placeholder?: string | undefined;
    /**
     * Styles to apply to the parent <View>.
     */
    containerStyle?: StyleProp<ViewStyle> | undefined;
    /**
     * Whether to show the cancel button while the user is typing.
     */
    showCancel?: boolean | undefined;
    /**
     * A callback that gets called when the Cancel button is pressed. It receives no arguments.
     */
    onCancel?: () => void;
}

/**
 * A component that acts as a regular <TextInput> but is styled to look like a SearchBar.
 */
const SearchBar = forwardRef((
    {
        placeholder,
        clearButtonMode,
        containerStyle,
        placeholderTextColor,
        onFocus,
        onBlur,
        showCancel,
        onCancel,
        style,
        ...props
    }: SearchBarProps,
    ref: React.MutableRefObject<TextInput>
) => {
    const colors = useColors();
    const backgroundColor = colors.COLOR_SCHEME == "light" ? "rgba(118, 118, 128, 0.12)" : "rgba(118, 118, 128, 0.24)";
    placeholderTextColor = typeof placeholderTextColor == "string" ? placeholderTextColor : colors.SECONDARY_TEXT;
    clearButtonMode = (
        clearButtonMode == "never" ||
        clearButtonMode == "while-editing" ||
        clearButtonMode == "unless-editing" ||
        clearButtonMode == "always"
    ) ? clearButtonMode : "always";

    showCancel = typeof showCancel == "boolean" ? showCancel : typeof onCancel == "function";

    const [isFocused, setIsFocused] = useState(false);
    const onFocusWrapped = useCallback(function(e: NativeSyntheticEvent<TextInputFocusEventData>) {
        setIsFocused(true);
        typeof onFocus == "function" ? onFocus(e) : null;
    }, [onFocus]);
    const onBlurWrapped = useCallback(function(e: NativeSyntheticEvent<TextInputFocusEventData>) {
        setIsFocused(false);
        typeof onBlur == "function" ? onBlur(e) : null;
    }, [onBlur]);

    ref = ref || useRef();

    return (
        <View style={[
            {
                display: "flex",
                flexDirection: "row",
                alignItems: "center"
            },
            containerStyle
        ]}>
            <View
                style={{
                    position: "absolute",
                    zIndex: 1,
                    height: "100%",
                    justifyContent: "center",
                    alignContent: "center",
                    paddingVertical: 0,
                    paddingLeft: 8,
                    paddingRight: 6,
                    width: 8 + 6 + 16
                }}
            >
                <Icon
                    width={16}
                    color={colors.SECONDARY_TEXT}
                >Search</Icon>
            </View>
            <TextInput
                ref={ref}
                style={[
                    {
                        backgroundColor,
                        borderRadius: style_values.border_radius,
                        paddingRight: 8,
                        paddingLeft: 8 + 6 + 16,
                        paddingVertical: 7,
                        fontSize: 16,
                        color: colors.TEXT,
                        display: "flex",
                        flexGrow: 1,
                        flexShrink: 1
                    },
                    style
                ]}
                clearButtonMode={clearButtonMode}
                placeholder={placeholder}
                placeholderTextColor={placeholderTextColor}
                onFocus={onFocusWrapped}
                onBlur={onBlurWrapped}
                {...props}
            />
            {
                showCancel && isFocused ?
                <TouchableOpacity
                    onPress={e => {
                        if (typeof onCancel == "function") {
                            onCancel();
                        }
                    }}
                >
                    <StyledText style={{fontSize: 16, textDecorationLine: "underline", marginLeft: 10}}>Cancel</StyledText>
                </TouchableOpacity> :
                null
            }
        </View>
    )
});

export default SearchBar;