import React, {forwardRef} from "react";
import { Text, TextProps } from "react-native";
import useColors from "../scripts/useColors";

interface StyledTextProps extends TextProps {
    /**
     * The type of text to display. This affects the text's styling, like font size and color, but
     * those styles can be overridden by supplying your own `style` prop. Default is `"primary"`.
     */
    type?: "primary" | "secondary" | "title" | "subtitle" | "h1" | "h2" | "h3" | undefined;
};

/**
 * A generic text component. Its text color is automatically adapted to the color theme
 * (`useColors().TEXT`) and its font-size is initialized to 16. Both of these styles can be over-
 * ridden by supplying your own fontSize and color attributes in the style prop.
 */
export const ColoredText = forwardRef((props, ref) => {
    const colors = useColors();
    return (
        <Text
            {...props}
            ref={ref}
            style={[
                {
                    color: colors.TEXT,
                    fontSize: 16,
                },
                props.style,
            ]}
        >
            {props.children}
        </Text>
    );
});

/**
 * Text that changes its color according to the current color theme (i.e., changes automatically
 * when the app switches between light and dark mode).
 */
const StyledText = ({type, style, ...props}: StyledTextProps) => {
    if (typeof type != "string") {
        type = "primary";
    }

    const colors = useColors();
    const styles = {
        fontSize: 16,
        color: colors.PRIMARY_TEXT
    };

    let fontWeight = null;
    let fontStyle = null;

    const passedStyles = style && typeof style == "object" ? style : {};
    if ("fontWeight" in passedStyles) {
        if (typeof passedStyles.fontWeight == "string" && passedStyles.fontWeight == "normal") {
            fontWeight = 400;
        } else if (typeof passedStyles.fontWeight == "string" && passedStyles.fontWeight == "bold") {
            fontWeight = 700;
        } else if ((typeof passedStyles.fontWeight == "number" || typeof passedStyles.fontWeight == "string") && 1 <= passedStyles.fontWeight && passedStyles.fontWeight < 450) {
            fontWeight = 400
        } else if ((typeof passedStyles.fontWeight == "number" || typeof passedStyles.fontWeight == "string") && 450 <= passedStyles.fontWeight && passedStyles.fontWeight < 600) {
            fontWeight = 500
        } else if ((typeof passedStyles.fontWeight == "number" || typeof passedStyles.fontWeight == "string") && 600 <= passedStyles.fontWeight && passedStyles.fontWeight <= 1000) {
            fontWeight = 700
        }
        delete passedStyles.fontWeight;
    }
    if ("fontStyle" in passedStyles) {
        if (typeof passedStyles.fontStyle == "string" && passedStyles.fontStyle == "normal") {
            fontStyle = "normal";
        } else if (typeof passedStyles.fontStyle == "string" && passedStyles.fontStyle == "italic") {
            fontStyle = "italic";
        }
        delete passedStyles.fontStyle;
    }

    switch (type) {
        default:
        case "primary":
            fontWeight = fontWeight === null ? 400 : fontWeight;
            fontStyle = fontStyle === null ? "normal" : fontStyle;
            break;
        case "secondary":
            fontWeight = fontWeight === null ? 400 : fontWeight;
            fontStyle = fontStyle === null ? "normal" : fontStyle;
            styles.color = colors.SECONDARY_TEXT;
            break;
        case "h1":
        case "title":
            fontWeight = fontWeight === null ? 500 : fontWeight;
            fontStyle = fontStyle === null ? "normal" : fontStyle;
            styles.fontSize = 30;
            break;
        case "h2":
        case "subtitle":
            fontWeight = fontWeight === null ? 700 : fontWeight;
            fontStyle = fontStyle === null ? "normal" : fontStyle;
            styles.fontSize = 20;
            styles.textTransform = "uppercase";
            styles.letterSpacing = 20 * 0.1;
            break;
        case "h3":
            fontWeight = fontWeight === null ? 500 : fontWeight;
            fontStyle = fontStyle === null ? "normal" : fontStyle;
            styles.textTransform = "uppercase";
            break
    }

    styles.fontFamily = `DMSans_${fontWeight == 700 ? "700Bold" : fontWeight == 500 ? "500Medium" : "400Regular"}${fontStyle == "italic" ? "_Italic" : ""}`;

    return <Text style={[styles, passedStyles]} {...props}/>
};

export default StyledText;