import React, {useState, useEffect, useRef, useCallback, useMemo } from "react";
import { View, ScrollView, Image, Dimensions, Animated, Easing, Alert, Platform, Pressable } from "react-native";
import StyledText from "../components/StyledText";
import usePaddedSafeAreaInsets from "../scripts/usePaddedSafeAreaInsets";
import distanceBetweenPoints from "../scripts/distanceBetweenPoints";
import getSelfLocation from "../scripts/getSelfLocation";
import useColors from "../scripts/useColors";
import { LinearGradient } from "expo-linear-gradient";
import { approximateLocations, selfUser, user_pool } from "../scripts/hardcoded_data";
import RowButton from "../components/RowButton";
import MaskedView from "@react-native-masked-view/masked-view";
import Svg, { Defs, Mask, RadialGradient, Rect, Stop } from "react-native-svg";
import { style_values } from "../scripts/styles";
import GlobalEvents from "../scripts/GlobalEvents";
import BackBtn from "../components/BackBtn";
import InterestList from "../components/InterestList";
import { Animated as AnimatedMapView, AnimatedRegion, MarkerAnimated } from "react-native-maps";
import useTriggerRerender from "../scripts/useTriggerRerender";
import * as Location from "expo-location";
import Icon from "../components/Icon";
import { BlurView } from "expo-blur";

/**
 * The maximum number of attempts to query for the other person's location. If the user reaches this
 * many attempts without getting a proper response, we assume they are no longer connected to the
 * internet and show an error message.
 * @constant
 * @type {Number}
 * @default 3
 */
const MAX_FAILED_ATTEMPTS = 3;

/**
 * The number of milliseconds that have to pass before we re-query for the other user's location.
 * @constant
 * @type {Number}
 * @default 2000
 */
const QUERY_INTERVAL = 2000;

const CollisionLocationSharingScreen = (props) => {
    const insets = usePaddedSafeAreaInsets();
    const colors = useColors();
    /** @type {import("@react-navigation/native").NavigationProp<ReactNavigation.RootParamList>} */
    const navigation = props.navigation;
    /** @type {import("../scripts/hardcoded_data").UserID} */
    const userID = props.route.params.userID;

    const hasSentLocation = useRef(true);
    const hasReceivedLocation = useRef(true);

    let distance = distanceBetweenPoints(
        getSelfLocation().coords,
        approximateLocations[userID].coords,
        "mi"
    );
    distance = Math.max(0.1, distance >= 10 ? Math.round(distance) : Math.round(distance * 1e1) / 1e1) + " mi";

    const askStopSharing = useCallback(() => {
        Alert.alert(
            "Stop Sharing Location?",
            `Are you sure you want to stop sharing your location with ${user_pool[userID].firstName}? You and ${user_pool[userID].firstName} will no longer have access to each other's locations.`,
            [
                {
                    text: "Stop Sharing",
                    style: "destructive",
                    onPress: () => {
                        GlobalEvents.emit("collisionsendlocation", false);
                    }
                },
                {
                    text: "Cancel",
                    style: "cancel"
                }
            ],
            {
                cancelable: true
            }
        );
    }, []);

    // Go back when either the self user or the other stop sharing their location.
    GlobalEvents.useListener("collisionreceivelocation", (received) => {
        hasReceivedLocation.current = received;
        if (!hasReceivedLocation.current || !hasSentLocation.current) {
            navigation.goBack();
        }
    }, []);
    GlobalEvents.useListener("collisionsendlocation", (sent) => {
        hasSentLocation.current = sent;
        if (!hasReceivedLocation.current || !hasSentLocation.current) {
            navigation.goBack();
        }
    }, []);

    const commonInterests = useMemo(() => {
        const commonInterests = [];
        const larger = new Set(selfUser.interests.length < user_pool[userID].interests.length ? user_pool[userID].interests : selfUser.interests);
        const smaller = selfUser.interests.length < user_pool[userID].interests.length ? selfUser.interests : user_pool[userID].interests;
        for (const interest of smaller) {
            if (larger.has(interest)) commonInterests.push(interest);
        }
        return commonInterests;
    }, []);

    return (
        <View
            style={{
                backgroundColor: colors.BACKGROUND,
                height: "100%"
            }}
        >
            <ScrollView
                contentContainerStyle={{
                    paddingTop: insets.top,
                    paddingBottom: insets.bottom,
                    paddingLeft: insets.left,
                    paddingRight: insets.right,
                }}
            >
                <MaskedView
                    style={{
                        position: "absolute",
                        left: -insets.left,
                        top: -insets.top,
                        right: -insets.right,
                        height: 600
                    }}
                    maskElement={
                        <LinearGradient
                            style={{
                                height: "100%",
                                width: "100%"
                            }}
                            colors={["rgba(0,0,0,1)", "rgba(0,0,0,.7)", "rgba(0,0,0,0)"]}
                            locations={[.6,.8,1]}
                        />
                    }
                >
                    <LinearGradient
                        style={{
                            height: "100%",
                            width: "100%"
                        }}
                        colors={[colors.FLAT2, colors.FLAT1]}
                        locations={[0, .5]}
                    />
                </MaskedView>
                <View
                    style={{
                        aspectRatio: 5/4,
                        width: "100%",
                        justifyContent: "center",
                        alignItems: "center"
                    }}
                >
                    <Svg
                        style={{
                            position: "absolute",
                            top: 0,
                            left: 0,
                            height: "100%",
                            width: "100%"
                        }}
                        viewBox="0 0 5 4"
                    >
                        <Defs>
                            <RadialGradient
                                id="gradient"
                            >
                                <Stop
                                    stopColor={colors.INVERTED_TEXT}
                                    stopOpacity={0.75}
                                    offset={.3}
                                />
                                <Stop
                                    stopColor={colors.INVERTED_TEXT}
                                    stopOpacity={0.7}
                                    offset={.4}
                                />
                                <Stop
                                    stopColor={colors.INVERTED_TEXT}
                                    stopOpacity={0}
                                    offset={1}
                                />
                            </RadialGradient>
                        </Defs>
                        <Rect
                            x={0}
                            y={0}
                            width={5}
                            height={4}
                            fill="url(#gradient)"
                        />
                    </Svg>
                    <ProfileBubbles
                        ID1={selfUser.id}
                        ID2={userID}
                    />
                </View>
                <StyledText type="h2" style={{
                    marginTop: 10,
                    fontSize: 40,
                    textAlign: "center",
                    paddingHorizontal: 10,
                    letterSpacing: 4,
                    color: colors.INVERTED_TEXT
                }}>You Collided!</StyledText>
                <StyledText type="h2" style={{
                    textAlign: "center",
                    marginTop: 50,
                    marginBottom: 5
                }}>Your Similarities</StyledText>
                <InterestList
                    similar={true}
                    interests={commonInterests}
                />
                <MapSection
                    style={{ marginTop: 40 }}
                    selfID={selfUser.id}
                    otherID={userID}
                />
                <View style={{ alignItems: "stretch", marginTop: 35 }}>
                    <RowButton
                        onPress={askStopSharing}
                        style={{
                            marginTop: 10,
                            marginBottom: 10
                        }}
                    >Stop Sharing Location</RowButton>
                    <RowButton
                        onPress={e => {
                            navigation.navigate("Other's Profile Screen", { userID, distance });
                        }}
                        style={{
                            marginBottom: 10
                        }}
                    >View Profile</RowButton>
                    <RowButton
                        onPress={e => {
                            navigation.navigate("Messaging Screen", { userID })
                        }}
                    >Chat</RowButton>
                </View>
            </ScrollView>
            <BackBtn
                onPress={askStopSharing}
            />
        </View>
    );
};

/**
 * @param {object} props
 * @param {import("../scripts/hardcoded_data").UserID} props.ID1
 * @param {import("../scripts/hardcoded_data").UserID} props.ID2
 */
const ProfileBubbles = ({ID1, ID2, props}) => {
    const [totalSize, setTotalSize] = useState((Dimensions.get("window").width - style_values.screen_padding.horizontal * 2) * .8);
    const colors = useColors("light");

    const BUBBLE_BOB_DURATION = 1500;

    const bubblePosition = useRef(new Animated.Value(-1)).current;

    useEffect(() => {
        const animation = Animated.loop(
            Animated.sequence([
                Animated.timing(
                    bubblePosition,
                    {
                        toValue: 1,
                        duration: BUBBLE_BOB_DURATION / 2,
                        useNativeDriver: true,
                        isInteraction: false,
                        easing: Easing.inOut(Easing.ease)
                    }
                ),
                Animated.timing(
                    bubblePosition,
                    {
                        toValue: -1,
                        duration: BUBBLE_BOB_DURATION / 2,
                        useNativeDriver: true,
                        isInteraction: false,
                        easing: Easing.inOut(Easing.ease)
                    }
                )
            ])
        );
        animation.start();
        return () => void animation.stop();
    }, []);

    const leftBubbleTransform = [
        { translateX: -Math.cos(Math.PI / 4) * (3 / 14 * totalSize - 8) },
        { translateY: +Math.sin(Math.PI / 4) * (3 / 14 * totalSize - 8) }
    ];
    const rightBubbleTransform = [
        { translateX: +Math.cos(Math.PI / 4) * (3 / 14 * totalSize - 8) },
        { translateY: -Math.sin(Math.PI / 4) * (3 / 14 * totalSize - 8) }
    ];

    // The borders for the collided bubbles have to be calculated using multiple <MaskedView>s, so
    // the code here is pretty gross.

    return (
        <Animated.View
            onLayout={({nativeEvent: {layout: {width}}}) => {
                setTotalSize(width);
            }}
            style={{
                aspectRatio: 1/1,
                height: "100%",
                justifyContent: "center",
                alignItems: "center",
                transform: [
                    { translateY: Animated.multiply(bubblePosition, totalSize / 50) }
                ]
            }}
        >
            {/* Left bubble (the one with the self user). */}
            <View style={{
                position: "absolute",
                aspectRatio: 1/1,
                height: 6 / 14 * totalSize,
                borderRadius: totalSize,
                overflow: "hidden",
                transform: leftBubbleTransform
            }}>
                <LinearGradient
                    style={{
                        height: "100%",
                        width: "100%",
                        padding: 6
                    }}
                    colors={[colors.FLAT2, colors.FLAT1]}
                    start={{ x: 0.211324865405, y: 1 }}
                    end={{ x: 0.788675134595, y: 0 }}
                >
                    <View style={{
                        height: "100%",
                        width: "100%",
                        borderRadius: totalSize,
                        backgroundColor: colors.BACKGROUND,
                        padding: 4
                    }}>
                        <Image
                            style={{
                                height: "100%",
                                width: "100%",
                                borderRadius: totalSize
                            }}
                            source={{
                                uri: user_pool[ID1].profileImage
                            }}
                        />
                    </View>
                </LinearGradient>
            </View>
            {/* The right bubble (the one with the right user). */}
            <View style={{
                position: "absolute",
                aspectRatio: 1/1,
                height: 6 / 14 * totalSize,
                borderRadius: totalSize,
                overflow: "hidden",
                transform: rightBubbleTransform
            }}>
                <LinearGradient
                    style={{
                        height: "100%",
                        width: "100%",
                        padding: 6
                    }}
                    colors={[colors.FLAT3, colors.FLAT4]}
                    start={{ x: 0.211324865405, y: 0 }}
                    end={{ x: 0.788675134595, y: 1 }}
                >
                    <View style={{
                        height: "100%",
                        width: "100%",
                        borderRadius: totalSize,
                        backgroundColor: colors.BACKGROUND,
                        padding: 4
                    }}>
                        <Image
                            style={{
                                height: "100%",
                                width: "100%",
                                borderRadius: totalSize
                            }}
                            source={{
                                uri: user_pool[ID2].profileImage
                            }}
                        />
                    </View>
                </LinearGradient>
            </View>
            {/* A jumble of <MaskedView> intersections to have the border of the bubbles overlap in
                an interlocked-looking way. */}
            <MaskedView
                style={{
                    position: "absolute",
                    top: 0,
                    left: 0,
                    height: "100%",
                    width: "100%",
                    justifyContent: "center",
                    alignItems: "center"
                }}
                maskElement={
                    <View style={{
                        height: "100%",
                        width: "100%",
                        justifyContent: "center",
                        alignItems: "center"
                    }}>
                        <View style={{
                            position: "absolute",
                            aspectRatio: 1/1,
                            height: 6 / 14 * totalSize - 12,
                            borderRadius: totalSize,
                            backgroundColor: "black",
                            transform: rightBubbleTransform
                        }}/>
                        <MaskedView
                            style={{
                                position: "absolute",
                                top: 0,
                                left: 0,
                                height: "100%",
                                width: "100%",
                                justifyContent: "center",
                                alignItems: "center"
                            }}
                            maskElement={
                                <View style={{
                                    height: "100%",
                                    width: "100%",
                                    justifyContent: "center",
                                    alignItems: "center"
                                }}>
                                    <View style={{
                                        borderWidth: 6,
                                        borderColor: "black",
                                        height: 6 / 14 * totalSize,
                                        width: 6 / 14 * totalSize,
                                        borderRadius: totalSize,
                                        transform: leftBubbleTransform
                                    }}/>
                                </View>
                            }
                        >
                            <View style={{
                                position: "absolute",
                                height: 3 / 14 * totalSize,
                                width: 6 / 14 * totalSize,
                                backgroundColor: "black",
                                transform: rightBubbleTransform.concat([
                                    { rotateZ: "-45deg" },
                                    { translateY: 1.5 / 14 * totalSize }
                                ])
                            }}/>
                        </MaskedView>
                    </View>
                }
            >
                <View style={{
                    position: "absolute",
                    aspectRatio: 1/1,
                    height: 6 / 14 * totalSize,
                    borderRadius: totalSize,
                    overflow: "hidden",
                    transform: leftBubbleTransform
                }}>
                    <LinearGradient
                        style={{
                            height: "100%",
                            width: "100%",
                            padding: 6
                        }}
                        colors={[colors.FLAT2, colors.FLAT1]}
                        start={{ x: 0.211324865405, y: 1 }}
                        end={{ x: 0.788675134595, y: 0 }}
                    >
                        <View style={{
                            height: "100%",
                            width: "100%",
                            borderRadius: totalSize,
                            backgroundColor: colors.BACKGROUND
                        }}/>
                    </LinearGradient>
                </View>
            </MaskedView>
        </Animated.View>
    );
};

/**
 * Returns a Region that encompasses all of the passed in coordinates.
 * @param {Iterable<import("expo-location").LocationObjectCoords} coordinates
 * @param {Number} [minimumDelta] The minimum value to use as the latitude and longitude delta.
 *      Default is 0.0001.
 * @param {Number|null} [aspectRatio] The aspect ratio that the `longitudeDelta` have with
 *      `latitudeDelta`. This is useful for if the map is of a certain aspect ratio and the
 *      latitude and longitude deltas need to match. If omitted or null, no aspect ratio is used and
 *      the deltas are returned as they are calculated. The ratio is width/height.
 * @return {import("react-native-maps").Region}
 */
const calculateRegion = (coordinates, minimumDelta, aspectRatio) => {
    minimumDelta = typeof minimumDelta == "number" && (isFinite(minimumDelta) || minimumDelta < 0) ? Math.max(minimumDelta, 0) : 0.0001;
    aspectRatio = typeof aspectRatio == "number" && aspectRatio > 0 && isFinite(aspectRatio) ? aspectRatio : null;

    let minLatitude = Infinity;
    let maxLatitude = -Infinity;
    let minLongitude = Infinity;
    let maxLongitude = -Infinity;

    for (const coordinate of coordinates) {
        minLatitude = Math.min(coordinate.latitude, minLatitude);
        maxLatitude = Math.max(coordinate.latitude, maxLatitude);
        minLongitude = Math.min(coordinate.longitude, minLongitude);
        maxLongitude = Math.max(coordinate.longitude, maxLongitude);
    }

    let latitudeDelta = Math.max((maxLatitude - minLatitude) * 1.4, minimumDelta);
    let longitudeDelta = Math.max((maxLongitude - minLongitude) * 1.4, minimumDelta);

    if (aspectRatio !== null) {
        if (latitudeDelta * aspectRatio < longitudeDelta) {
            latitudeDelta = longitudeDelta / aspectRatio;
        } else {
            longitudeDelta = latitudeDelta * aspectRatio;
        }
    }

    return {
        latitude: (minLatitude + maxLatitude) / 2,
        longitude: (minLongitude + maxLongitude) / 2,
        latitudeDelta,
        longitudeDelta
    };
};

/**
 * @param {object} props
 * @param {import("../scripts/hardcoded_data").UserID} props.selfID
 * @param {import("../scripts/hardcoded_data").UserID} props.otherID
 * @param {import("react-native").StyleProp<import("react-native").ViewStyle>} props.style
 */
const MapSection = ({selfID, otherID, style, ...props}) => {
    const colors = useColors();
    // The user shouldn't be able to navigate this far if we don't have location permission (and
    // really shouldn't be able to collide with anyone in the first place), but this just doubly
    // makes sure we have location permissions.
    Location.useForegroundPermissions();

    /** @type {React.MutableRefObject<import("expo-location").LocationObject>} */
    const selfLocation = useRef({});
    /** @type {React.MutableRefObject<import("expo-location").LocationObject>} */
    const otherLocation = useRef({});
    const queryAttempts = useRef(0);
    const triggerRerender = useTriggerRerender();

    // Calculate the aspect ratio of the map so that it only takes up a maximum of 40% of the screen
    // height. This is because the user still needs area on the screen to be able to scroll (since
    // scrolling along the map would just move the map instead). The tallest it will get is having
    // an aspect ratio of 1/1.
    const windowAspectRatio = Dimensions.get("window").width / Dimensions.get("window").height;
    const mapAspectRatio = Math.max(1, windowAspectRatio * 2.5);

    const isLoading = !selfLocation.current.coords || !otherLocation.current.coords;

    // A state indicating whether the user is having connectivity problems. For now, nothing happens
    // when this is true, but in the final version, this should show an error message of some kind
    // and have options for the user to proceed.
    const [hasConnectivityProblems, setHasConnectivityProblems] = useState(false);

    // The region that gets shown in the map as an animated value.
    const mapRegion = useRef(new AnimatedRegion()).current;

    // Whether the map is currently tracking the pins. If this is true, the region will update
    // automatically when the pins move.
    const isTracking = useRef(true);

    const onRegionChange = useCallback((newRegion) => {
        mapRegion.setValue(newRegion);
    }, []);

    // Whether this component has already unmounted.
    const hasUnmounted = useRef(false);

    // Remembers the previous distance. This is used to determine the accuracy level of our
    // location. We start off with a pretty bad accuracy, and then as the users get closer, their
    // location accuracy will improve to give them better directions, but at the cost of taking
    // longer to fetch their locations.
    const previousDistance = useRef(1000);

    const queryLocations = useCallback(() => {
        Promise.allSettled([
            Location.getCurrentPositionAsync({
                accuracy:
                    previousDistance.current > 300 ? Location.Accuracy.Balanced :
                    previousDistance.current > 50 ? Location.Accuracy.High :
                    Location.Accuracy.Highest
            }),
            queryOthersDistance(selfID, otherID)
        ]).then(
            results => {
                // Do nothing if the component has already unmounted.
                if (hasUnmounted.current) return;

                if (results[0].status == "fulfilled") {
                    selfLocation.current = results[0].value;
                    triggerRerender();
                }

                if (results[1].status == "fulfilled") {
                    otherLocation.current = results[1].value;

                    if (Platform.OS === "android") {
                        otherPositionRef.current?.animateMarkerToCoordinate(
                            {
                                latitude: results[1].value.coords.latitude,
                                longitude: results[1].value.coords.longitude
                            },
                            300
                        );
                    } else {
                        otherPositionAnimatedRegion.timing({
                            latitude: results[1].value.coords.latitude,
                            longitude: results[1].value.coords.longitude,
                            useNativeDriver: true,
                            duration: 300
                        }).start();
                    }

                    queryAttempts.current = 0;
                    triggerRerender();
                    setHasConnectivityProblems(false);
                } else {
                    queryAttempts.current++;
                }

                // If we have reached the maximum number of attempts, we can assume the user just
                // doesn't have a strong internet connection.
                if (queryAttempts.current >= MAX_FAILED_ATTEMPTS) {
                    setHasConnectivityProblems(true);
                }

                // Query retrieving distances again
                setTimeout(queryLocations, QUERY_INTERVAL);
            },
            timeoutRejection => {
                // Do nothing of the component has already unmounted.
                if (hasUnmounted.current) return;

                setHasConnectivityProblems(true);
            }
        );
    }, []);
    // Start querying for locations.
    useEffect(() => {
        queryLocations();
        return () => void (hasUnmounted.current = true);
    }, []);

    // The AnimatedRegion controlling the pin for the other user.
    const otherPositionAnimatedRegion = useRef(new AnimatedRegion()).current;
    // A ref to the <MarkerAnimated> controlling the other user.
    /** @type {React.MutableRefObject<MarkerAnimated>} */
    const otherPositionRef = useRef();

    // Memoize the Map component so it doesn't have to re-render every time.
    const memoizedMap = useMemo(() => {
        if (isLoading) return null;
        return (
            <AnimatedMapView
                showsUserLocation={true}
                loadingEnabled={true}
                showsMyLocationButton={false}
                userInterfaceStyle={colors.COLOR_SCHEME}
                onPanDrag={e => {
                    if (isTracking.current) {
                        isTracking.current = false;
                        triggerRerender();
                    }
                }}
                style={{
                    height: "100%",
                    width: "100%"
                }}
                region={mapRegion}
                onRegionChange={onRegionChange}
            >
                <MarkerAnimated
                    ref={otherPositionRef}
                    coordinate={otherPositionAnimatedRegion}
                />
            </AnimatedMapView>
        );
    }, [isLoading, colors.COLOR_SCHEME]);

    const memoizedAccuracyTimer = useMemo(() => {
        return <AccuracyTimer
            timestamps={[
                selfLocation.current.timestamp,
                otherLocation.current.timestamp
            ]}
        />
    }, [selfLocation.current?.timestamp, otherLocation.current?.timestamp]);

    const [isFilledLocationBeingPressed, setIsFilledLocationBeingPressed] = useState(false);

    // If we're still loading the page, return a generic, empty view.
    if (isLoading) {
        return (
            <View style={style}>
                <StyledText
                    type="h2"
                    style={{
                        textAlign: "center"
                    }}
                >Getting locations...</StyledText>
                <View style={{
                    aspectRatio: mapAspectRatio,
                    width: "100%",
                    borderRadius: style_values.border_radius,
                    overflow: "hidden",
                    backgroundColor: colors.SECONDARY_TEXT,
                    opacity: 0.5
                }}/>
            </View>
        );
    }

    let distance =distanceBetweenPoints(
        selfLocation.current.coords,
        otherLocation.current.coords,
        "feet"
    );
    previousDistance.current = distance;
    // Round distance to multiples of 5.
    distance = distance < 5 ? "<5 feet" : Math.round(distance / 5) * 5 + " feet";

    if (isTracking.current) {
        const region = calculateRegion(
            [selfLocation.current.coords, otherLocation.current.coords],
            null,
            mapAspectRatio
        );
        mapRegion.timing({
            latitude: region.latitude,
            longitude: region.longitude,
            latitudeDelta: region.latitudeDelta,
            longitudeDelta: region.longitudeDelta,
            duration: 300,
            useNativeDriver: true
        }).start();
    }

    return (
        <View style={style}>
            <StyledText
                type="h2"
                style={{
                    textAlign: "center",
                    marginBottom: 5
                }}
            >{user_pool[otherID].firstName} is {distance} away</StyledText>
            <View style={[
                {
                    aspectRatio: mapAspectRatio,
                    width: "100%",
                    borderRadius: style_values.border_radius,
                    overflow: "hidden"
                }
            ]}>
                {memoizedMap}
                <Pressable
                    pointerEvents={isTracking.current ? "none" : "auto"}
                    style={{
                        position: "absolute",
                        top: 10,
                        left: 10,
                        borderRadius: style_values.border_radius,
                        overflow: "hidden",
                        padding: 8,
                        opacity: isTracking.current ? 0 : 1
                    }}
                    onPressIn={e => setIsFilledLocationBeingPressed(true)}
                    onPressOut={e => setIsFilledLocationBeingPressed(false)}
                    onPress={e => {
                        isTracking.current = true;
                        triggerRerender();
                    }}
                >
                    <BlurView
                        style={{
                            position: "absolute",
                            top: 0,
                            right: 0,
                            bottom: 0,
                            left: 0
                        }}
                        intensity={65}
                        tint={colors.COLOR_SCHEME}
                    />
                    <Icon height={26} color={isFilledLocationBeingPressed ? colors.FLAT1 : colors.TEXT}>fluent:my-location-filled</Icon>
                </Pressable>
            </View>
            {memoizedAccuracyTimer}
        </View>
    )
};

/**
 * @param {object} props
 * @param {Array<Number>} props.timestamps The timestamps to use for calculating accuracy. Each
 *      timestamp should be the number of milliseconds since the epoch.
 */
const AccuracyTimer = ({timestamps, ...props}) => {
    const farthestTimestamp = Math.min(...timestamps);
    const [timeDeltaText, setTimeDeltaText] = useState("0 seconds");

    useEffect(() => {
        const interval = setInterval(() => {
            const now = new Date();
            const delta = now.valueOf() - farthestTimestamp;
            setTimeDeltaText(
                delta < 1 * 1000 ? "0 seconds" :
                delta < 2 * 1000 ? "1 second" :
                delta < 60 * 1000 ? `${Math.floor(delta / 1000)} seconds` :
                delta < 120 * 1000 ? "1 minute" :
                delta < 60 * 60 * 1000 ? `${Math.floor(delta / 60 / 1000)} minutes` :
                delta < 120 * 60 * 1000 ? "1 hour" :
                `${Math.floor(delta / 60 / 60 / 1000)} hours`
            );
        }, 50);
        return () => void clearInterval(interval);
    });

    return (
        <StyledText
            type="secondary"
            style={{
                textAlign: "center",
                fontSize: 14
            }}
        >Accurate as of {timeDeltaText} ago</StyledText>
    );
}

/**
 * @param {import("../scripts/hardcoded_data").UserID} selfID
 * @param {import("../scripts/hardcoded_data").UserID} userID
 * @returns {Promise<import("expo-location").LocationObject>}
 */
const queryOthersDistance = async (selfID, userID) => {
    const location = approximateLocations[userID];
    location.timestamp = new Date().valueOf();
    return location;
};

export default CollisionLocationSharingScreen;