import React, {useState, useEffect, useRef, useCallback} from "react";
import { View, ScrollView, Image, Dimensions, Animated, Easing, FlatList, TouchableOpacity } from "react-native";
import StyledText from "../components/StyledText";
import usePaddedSafeAreaInsets from "../scripts/usePaddedSafeAreaInsets";
import distanceBetweenPoints from "../scripts/distanceBetweenPoints";
import getSelfLocation, { ensureLocationPermission } from "../scripts/getSelfLocation";
import InterestBubble from "../components/InterestBubble";
import useColors from "../scripts/useColors";
import { LinearGradient } from "expo-linear-gradient";
import { approximateLocations, selfUser, user_pool, matchScores } from "../scripts/hardcoded_data";
import RowButton from "../components/RowButton";
import { useNavigation } from "@react-navigation/native";
import MaskedView from "@react-native-masked-view/masked-view";
import Svg, { Defs, RadialGradient, Rect, Stop } from "react-native-svg";
import { style_values } from "../scripts/styles";
import { BlurView } from "expo-blur";
import Icon from "../components/Icon";
import { Modalize } from 'react-native-modalize';

import GlobalEvents from "../scripts/GlobalEvents";
import useTriggerRerender from "../scripts/useTriggerRerender";

const CollisionScreen = (props) => {
    const insets = usePaddedSafeAreaInsets();
    const colors = useColors();
    /** @type {import("@react-navigation/native").NavigationProp<ReactNavigation.RootParamList>} */
    const navigation = props.navigation;
    /** @type {import("../scripts/hardcoded_data").UserID} */
    const userID = props.route.params.userID;

    const hasSentLocation = useRef(false);
    const hasReceivedLocation = useRef(false);
    const triggerRerender = useTriggerRerender();

    /** @type {import("react").MutableRefObject<Modalize>} */
    const modalizeRef = useRef();

    let distance = distanceBetweenPoints(
        getSelfLocation().coords,
        approximateLocations[userID].coords,
        "mi"
    );
    distance = Math.max(0.1, distance >= 10 ? Math.round(distance) : Math.round(distance * 1e1) / 1e1) + " mi";

    GlobalEvents.useListener("collisionreceivelocation", (received) => {
        hasReceivedLocation.current = received;
        triggerRerender();
        if (hasReceivedLocation.current && hasSentLocation.current) {
            navigation.navigate("Collision Location Sharing Screen", { userID });
        }
    }, []);

    GlobalEvents.useListener("collisionsendlocation", (sent) => {
        hasSentLocation.current = sent;
        triggerRerender();
        if (hasReceivedLocation.current && hasSentLocation.current) {
            navigation.navigate("Collision Location Sharing Screen", { userID });
        }
    }, []);

    useEffect(() => {
        const timeout = setTimeout(() => {
            GlobalEvents.emit("collisionreceivelocation", true);
        }, 4000);
        return () => clearInterval(timeout);
    }, []);

    return (
        <View
            style={{
                backgroundColor: colors.BACKGROUND,
                height: "100%"
            }}
        >
            <ScrollView
                contentContainerStyle={{
                    paddingTop: insets.top,
                    paddingBottom: insets.bottom,
                    paddingLeft: insets.left,
                    paddingRight: insets.right,
                }}
            >
                <InfoBtn modalizeRef={modalizeRef}/>
                <MaskedView
                    style={{
                        position: "absolute",
                        left: 0,
                        top: 0,
                        right: 0,
                        height: 800
                    }}
                    maskElement={
                        <View style={{
                            width: "100%",
                            paddingLeft: insets.left,
                            paddingRight: insets.right,
                            paddingTop: insets.top
                        }}>
                            <View
                                style={{
                                    aspectRatio: 5/4
                                }}
                            >
                                <View
                                    style={{
                                        position: "absolute",
                                        top: -insets.top,
                                        right: -insets.right,
                                        bottom: -200,
                                        left: -insets.left
                                    }}
                                >
                                    <LinearGradient
                                        style={{
                                            position: "absolute",
                                            top: 0,
                                            right: 0,
                                            bottom: 0,
                                            left: 0
                                        }}
                                        colors={["rgba(0,0,0,1)", "rgba(0,0,0,.7)", "rgba(0,0,0,0)"]}
                                        locations={[.6,.8,1]}
                                    />
                                </View>
                            </View>
                        </View>
                    }
                >
                    <View style={{
                        width: "100%",
                        paddingLeft: insets.left,
                        paddingRight: insets.right,
                        paddingTop: insets.top
                    }}>
                        <View
                            style={{
                                aspectRatio: 5/4
                            }}
                        >
                            <View
                                style={{
                                    position: "absolute",
                                    top: -insets.top,
                                    right: -insets.right,
                                    bottom: -200,
                                    left: -insets.left
                                }}
                            >
                                <LinearGradient
                                    style={{
                                        position: "absolute",
                                        top: 0,
                                        right: 0,
                                        bottom: 0,
                                        left: 0
                                    }}
                                    colors={[colors.FLAT2, colors.FLAT1]}
                                    locations={[0, .5]}
                                />
                            </View>
                        </View>
                    </View>
                </MaskedView>
                <View
                    style={{
                        aspectRatio: 5/4,
                        width: "100%",
                        justifyContent: "center",
                        alignItems: "center"
                    }}
                >
                    <Svg
                        style={{
                            position: "absolute",
                            top: 0,
                            left: 0,
                            height: "100%",
                            width: "100%",
                            opacity: 0.3
                        }}
                        viewBox="0 0 5 4"
                    >
                        <Defs>
                            <RadialGradient
                                id="gradient"
                            >
                                <Stop
                                    stopColor={colors.INVERTED_TEXT}
                                    stopOpacity={0.75}
                                    offset={.3}
                                />
                                <Stop
                                    stopColor={colors.INVERTED_TEXT}
                                    stopOpacity={0.7}
                                    offset={.4}
                                />
                                <Stop
                                    stopColor={colors.INVERTED_TEXT}
                                    stopOpacity={0}
                                    offset={1}
                                />
                            </RadialGradient>
                        </Defs>
                        <Rect
                            x={0}
                            y={0}
                            width={5}
                            height={4}
                            fill="url(#gradient)"
                        />
                    </Svg>
                    <ProfileBubbles
                        ID1={selfUser.id}
                        ID2={userID}
                    />
                </View>
                <StyledText type="h2" style={{
                    marginTop: 10,
                    fontSize: 40,
                    textAlign: "center",
                    paddingHorizontal: 10,
                    letterSpacing: 4,
                    color: colors.INVERTED_TEXT
                }}>You Collided!</StyledText>
                <StyledText type="h2" style={{
                    marginTop: 50,
                    fontSize: 30,
                    letterSpacing: 3,
                    textAlign: 'center'
                }}>Emily Hsu</StyledText>
                <MatchBar matchScore={matchScores[userID]} style={{ marginTop: 25}} />
                <StyledText type="h2" style={{
                    textAlign: "center",
                    marginTop: 50,
                    marginBottom: 5
                }}>Your Similarities</StyledText>
                <InterestBubblesGrid
                    interests1={selfUser.interests}
                    interests2={user_pool[userID].interests}
                    style={{
                        marginBottom: 40
                    }}
                />
                {
                    hasSentLocation.current && !hasReceivedLocation.current ?
                    <StyledText style={{textAlign: "center"}}>Waiting for {user_pool[userID].firstName} to share their location. Your location is not visible to {user_pool[userID].firstName} yet.</StyledText> :
                    !hasSentLocation.current && hasReceivedLocation.current ?
                    <StyledText style={{textAlign: "center"}}>{user_pool[userID].firstName} wants to exchange locations with you to meet in person. You can exchange your location below.</StyledText> :
                    null
                }
                <View style={{ alignItems: "stretch" }}>
                    {
                        hasSentLocation.current ?
                        <RowButton
                            onPress={e => {
                                GlobalEvents.emit("collisionsendlocation", false);
                            }}
                            style={{
                                marginTop: 10,
                                marginBottom: 10
                            }}
                        >Stop Sharing Location</RowButton> :
                        <RowButton
                            mainAction
                            onPress={e => {
                                ensureLocationPermission().then(
                                    hasPermission => {
                                        if (hasPermission) GlobalEvents.emit("collisionsendlocation", true);
                                    }
                                )
                            }}
                            style={{
                                marginTop: 10,
                                marginBottom: 10
                            }}
                        >Exchange Location</RowButton>
                    }
                    <RowButton
                        onPress={e => {
                            navigation.navigate("Other's Profile Screen", { userID, distance });
                        }}
                        style={{
                            marginBottom: 10
                        }}
                    >View Profile</RowButton>
                    <RowButton
                        onPress={e => {
                            navigation.navigate("Messaging Screen", { userID })
                        }}
                    >Chat</RowButton>
                </View>
            </ScrollView>
            <CloseBtn/>
            <Modalize ref={modalizeRef} modalTopOffset={200} childrenStyle={{backgroundColor: colors.BACKGROUND}}>
                <InfoScreen modalizeRef={modalizeRef}/>
            </Modalize>
        </View>
    );
};

/**
 * @param {object} props
 * @param {import("../scripts/hardcoded_data").UserID} props.ID1
 * @param {import("../scripts/hardcoded_data").UserID} props.ID2
 */
const ProfileBubbles = ({ID1, ID2, props}) => {
    const [totalSize, setTotalSize] = useState((Dimensions.get("window").width - style_values.screen_padding.horizontal * 2) * .8);
    const colors = useColors("light");

    const BUBBLE_BOB_DURATION = 1500;

    // This is an Animated.ValueXY that controls the position of the left bubble. The right bubble
    // is mirrored around the center so that its position is determined purely from the position of
    // the left bubble.
    const bubblePosition = useRef(new Animated.ValueXY({ x: -3 / 14, y: 1 / 4 })).current;

    useEffect(() => {
        const animation = Animated.loop(
            Animated.sequence([
                Animated.timing(
                    bubblePosition,
                    {
                        toValue: { x: -3 / 14, y: 5 / 28 },
                        duration: BUBBLE_BOB_DURATION / 2,
                        useNativeDriver: true,
                        isInteraction: false,
                        easing: Easing.inOut(Easing.ease)
                    }
                ),
                Animated.timing(
                    bubblePosition,
                    {
                        toValue: { x: -3 / 14, y: 3 / 28 },
                        duration: BUBBLE_BOB_DURATION / 2,
                        useNativeDriver: true,
                        isInteraction: false,
                        easing: Easing.inOut(Easing.ease)
                    }
                )
            ])
        );
        animation.start();
        return () => void animation.stop();
    }, []);

    return (
        <View
            onLayout={({nativeEvent: {layout: {width}}}) => {
                setTotalSize(width);
            }}
            style={{
                aspectRatio: 1/1,
                height: "100%",
                justifyContent: "center",
                alignItems: "center"
            }}
        >
            <Animated.View
                style={{
                    position: "absolute",
                    aspectRatio: 1/1,
                    height: 6 / 14 * 100 + "%",
                    borderRadius: totalSize,
                    overflow: "hidden",
                    transform: [
                        { translateX: Animated.multiply(bubblePosition.x, totalSize) },
                        { translateY: Animated.multiply(bubblePosition.y, totalSize) }
                    ]
                }}
            >
                <LinearGradient
                    style={{
                        height: "100%",
                        width: "100%",
                        padding: 6
                    }}
                    colors={[colors.FLAT2, colors.FLAT1]}
                    start={{ x: 0.211324865405, y: 1 }}
                    end={{ x: 0.788675134595, y: 0 }}
                >
                    <View
                        style={{
                            height: "100%",
                            width: "100%",
                            borderRadius: totalSize,
                            backgroundColor: colors.BACKGROUND,
                            padding: 4
                        }}
                    >
                        <Image
                            style={{
                                height: "100%",
                                width: "100%",
                                borderRadius: totalSize
                            }}
                            source={{
                                uri: user_pool[ID1].profileImage
                            }}
                        />
                    </View>
                </LinearGradient>
            </Animated.View>
            <Animated.View
                style={{
                    position: "absolute",
                    aspectRatio: 1/1,
                    height: 6 / 14 * 100 + "%",
                    borderRadius: totalSize,
                    overflow: "hidden",
                    transform: [
                        { translateX: Animated.multiply(bubblePosition.x, -totalSize) },
                        { translateY: Animated.multiply(bubblePosition.y, -totalSize) }
                    ]
                }}
            >
                <LinearGradient
                    style={{
                        height: "100%",
                        width: "100%",
                        padding: 6
                    }}
                    colors={[colors.FLAT3, colors.FLAT4]}
                    start={{ x: 0.211324865405, y: 0 }}
                    end={{ x: 0.788675134595, y: 1 }}
                >
                    <View
                        style={{
                            height: "100%",
                            width: "100%",
                            borderRadius: totalSize,
                            backgroundColor: colors.BACKGROUND,
                            padding: 4
                        }}
                    >
                        <Image
                            style={{
                                height: "100%",
                                width: "100%",
                                borderRadius: totalSize
                            }}
                            source={{
                                uri: user_pool[ID2].profileImage
                            }}
                        />
                    </View>
                </LinearGradient>
            </Animated.View>
        </View>
    );
};

/**
 * @param {object} props
 * @param {import("react-native").StyleProp<import("react-native").ViewStyle>} props.style
 * @param {Number} props.matchScore The score to display in the match bar. Should be a number in the
 *      range [0,1].
 */
const MatchBar = ({matchScore, style, ...props}) => {
    const colors = useColors();

    return (
        <View style={[
            {
                display: "flex",
                flexDirection: "column",
            },
            style
        ]}>
            <StyledText type="h2" style={{textAlign: "center"}}> 95% Match</StyledText>
            <View style={{
                borderWidth: 2,
                borderColor: colors.TEXT,
                borderRadius: 10,
                overflow: "hidden",
                height: 20
            }}>
                <LinearGradient
                    style={{
                        height: "100%",
                        width: Math.floor(matchScore * 100) + "%",
                    }}
                    colors={[colors.FLAT4, colors.FLAT3]}
                    start={[1.3, .5]}
                    end={[-.1, .5]}
                />
            </View>
        </View>
    )
};

/**
 * @param {object} props
 * @param {Number} props.blurIntensity
 */
const CloseBtn = ({blurIntensity, ...props}) => {
    const navigation = useNavigation();
    const colors = useColors();
    const insets = usePaddedSafeAreaInsets();
    blurIntensity = typeof blurIntensity == "number" && 0 <= blurIntensity && blurIntensity <= 100 ? blurIntensity : 65;

    const onPress = useCallback(e => {
        navigation.goBack();
    }, []);

    return (
        <View
            style={{
                position: "absolute",
                top: insets.top,
                right: insets.right,
                zIndex: 1,
                borderRadius: "50%",
                overflow: "hidden"
            }}
        >
            <TouchableOpacity
                style={{
                    height: "100%",
                    width: "100%"
                }}
                onPress={onPress}
            >
                <View
                    style={{
                        height: "100%",
                        width: "100%",
                        borderRadius: "50%",
                        overflow: "hidden"
                    }}
                >
                    <BlurView
                        tint={colors.COLOR_SCHEME}
                        intensity={blurIntensity}
                        style={{
                            padding: 12,
                            height: "100%",
                            width: "100%"
                        }}
                    >
                        <Icon height={20}>cross</Icon>
                    </BlurView>
                </View>
            </TouchableOpacity>
        </View>
    )
};

/**
 * @param {object} props
 * @param {import("react").MutableRefObject<Modalize>} props.modalizeRef
 */
const InfoScreen = ({modalizeRef, ...props}) => {
    const insets = usePaddedSafeAreaInsets();

    return(
        <View style={{
            paddingTop: style_values.screen_padding.vertical,
            paddingRight: insets.right,
            paddingBottom: insets.bottom,
            paddingLeft: insets.left
        }}>
            <View
                style={{
                    display: "flex",
                    flexDirection: "row",
                    justifyContent: "space-between",
                    alignItems: "center"
                }}
            >
                <StyledText style={{ fontWeight: 700, fontSize: 24, marginVertical: 4 }}>FAQ</StyledText>
                <TouchableOpacity
                    style={{
                        padding: 7
                    }}
                    onPress={e => {
                        modalizeRef.current?.close();
                    }}
                >
                    <Icon height={20}>cross</Icon>
                </TouchableOpacity>
            </View>

            <StyledText style={{ fontWeight: 700, marginVertical: 4 }}>What is a collision?</StyledText>
            <StyledText>There is someone nearby who you should get to know! You likely have a similar background or shared interests.</StyledText>

            <StyledText style={{ fontWeight: 700, marginTop: 12, marginBottom: 4 }}>What information is shared with my collision?</StyledText>
            <StyledText>We only share your public profile information, including your name, profile picture, and interests.</StyledText>

            <StyledText style={{ fontWeight: 700, marginTop: 12, marginBottom: 4 }}>Is my location being shared?</StyledText>
            <StyledText>We don't share your location with your collision unless you explicitly share your location by tapping <StyledText style={{fontWeight: "bold"}}>Exchange Location</StyledText>. You can always stop sharing your location at any time.</StyledText>

            <StyledText style={{ fontWeight: 700, marginTop: 12, marginBottom: 4 }}>Do I have to meet up with my collision?</StyledText>
            <StyledText>No, you don't have to! But if you want to meet up, you can message your collision, or exchange your location to help your collision find you.</StyledText>

            <StyledText style={{ fontWeight: 700, marginTop: 12, marginBottom: 4 }}>If I exit the collision, will my location still be shared?</StyledText>
            <StyledText>Once you exit the collision, your location will stop being shared. However, you can still re-enter the collision for the next 30 minutes as long as both you and your collision are active and within one mile of each other. You can access your collision from the home screen.</StyledText>
        </View>
    )
};

/**
 * @param {object} props
 * @param {import("react").MutableRefObject<Modalize>} props.modalizeRef
 */
const InfoBtn = ({modalizeRef, ...props}) => {
    const insets = usePaddedSafeAreaInsets();
    const colors = useColors();

    const onPress = useCallback(e => {
        modalizeRef.current?.open();
    }, []);

    return (
        <View
            style={{
                position: "absolute",
                zIndex: 100,
                padding: 7,
                top: insets.top,
                left: insets.left
            }}
        >
            <TouchableOpacity
                style={{
                    padding: 7,
                }}
                onPress={onPress}
            >
                <Icon height={30} color={colors.INVERTED_TEXT}>info</Icon>
            </TouchableOpacity>
        </View>
    );
};

/**
 * @param {object} props
 * @param {Array<String>} props.interests1
 * @param {Array<String>} props.interests2
 * @param {import("react-native").StyleProp<import("react-native").ViewStyle>} props.style
 * @param {import("react-native").StyleProp<import("react-native").ViewStyle>}
 *      props.contentContainerStyle
 */
const InterestBubblesGrid = ({interests1, interests2, style, contentContainerStyle, ...props}) => {
    const insets = usePaddedSafeAreaInsets();
    const commonInterests = [];
    const larger = new Set(interests1.length < interests2.length ? interests2 : interests1);
    const smaller = interests1.length < interests2.length ? interests1 : interests2;
    for (const interest of smaller) {
        if (larger.has(interest)) commonInterests.push(interest);
    }
    const MAX_ROWS = 4;
    const SPACING = 4;

    // If the number of items is less than the maximum number of rows squared, we display them in a
    // square pattern instead of adding on columns. It looks just a little bit nicer for smaller
    // number of interests.
    if (commonInterests.length <= MAX_ROWS * (MAX_ROWS + 1)) {
        const numRows = Math.ceil(Math.sqrt(commonInterests.length));
        const rows = new Array(Math.ceil(commonInterests.length / numRows));
        for (let i = 0; i < commonInterests.length; i++) {
            if (i % numRows == 0) rows[i / numRows] = new Array(Math.min(numRows, commonInterests.length - i));
            rows[Math.floor(i / numRows)][i % numRows] = commonInterests[i];
        }

        return (
            <ScrollView
                horizontal
                style={[
                    {
                        alignSelf: "center",
                        marginLeft: -insets.left,
                        marginRight: -insets.right
                    },
                    style
                ]}
                contentContainerStyle={[
                    {
                        display: "flex",
                        flexDirection: "column",
                        paddingLeft: insets.left,
                        paddingRight: insets.right
                    },
                    contentContainerStyle
                ]}
                {...props}
            >
                {
                    rows.map((row, index) =>
                        <React.Fragment key={row.join(",")}>
                            <View
                                style={{
                                    display: "flex",
                                    flexDirection: "row",
                                    justifyContent: "center"
                                }}
                            >{
                                row.map((interest, index) =>
                                    <React.Fragment key={interest}>
                                        <InterestBubble
                                            currInterestName={interest}
                                        />
                                        {
                                            index < row.length - 1 ?
                                            <View style={{ width: SPACING }}/> : null
                                        }
                                    </React.Fragment>
                                )
                            }</View>
                            {
                                index < rows.length - 1 ?
                                <View style={{ height: SPACING }}/> : null
                            }
                        </React.Fragment>
                    )
                }
            </ScrollView>
        )
    }

    // Assign each interest into a column. The ordering is still left to right despite the actual
    // columns being ordered top to bottom.
    /** @type Array<Array<any>> */
    const columns = new Array(Math.ceil(commonInterests.length / MAX_ROWS));
    for (let i = 0; i < commonInterests.length; i++) {
        if (i < columns.length) columns[i] = new Array(MAX_ROWS);
        columns[i % columns.length][Math.floor(i / columns.length)] = commonInterests[i];
    }

    return (
        <FlatList
            data={columns}
            horizontal
            keyExtractor={(item, index) => item.join(",")}
            renderItem={({item: column}) =>
                <View>{
                    column.map(interest =>
                        <InterestBubble
                            key={interest}
                            currInterestName={interest}
                        />
                    )
                }</View>
            }
            ItemSeparatorComponent={() => <View style={{width: SPACING}}/>}
            style={[
                {
                    marginLeft: -insets.left,
                    marginRight: -insets.right
                },
                style
            ]}
            contentContainerStyle={[
                {
                    paddingLeft: insets.left,
                    paddingRight: insets.right,
                    minWidth: "100%",
                    justifyContent: "center",
                    alignItems: "flex-start"
                },
                contentContainerStyle
            ]}
            {...props}
        />
    );
};

export default CollisionScreen;