import React, { useCallback, useState } from "react";
import { TouchableOpacity, Image, View, KeyboardAvoidingView, StyleSheet, FlatList } from "react-native";
import { LinearGradient } from "expo-linear-gradient";
import usePaddedSafeAreaInsets from "../scripts/usePaddedSafeAreaInsets";
import { messages_objects, selfUser, user_pool } from "../scripts/hardcoded_data";
import useColors from "../scripts/useColors";
import { useNavigation } from '@react-navigation/native';
import Icon from '../components/Icon';
import StyledText from "../components/StyledText";
import { style_values } from "../scripts/styles";
import Color from "color";
import StyledTextInput from "../components/StyledTextInput";
import MaskedView from "@react-native-masked-view/masked-view";

const MessagingScreen = (props) => {
    const insets = usePaddedSafeAreaInsets();

    const colors = useColors();
    const userID = props.route.params.userID;

    const [messages, setMessages] = useState(messages_objects[userID].slice());

    /** @type {(message: import("../scripts/hardcoded_data").Message) => void} */
    const sendMessage = useCallback((message) => {
        messages_objects[userID].push(message);
        setMessages([
            ...messages,
            message
        ]);
    },  [messages]);

    return (
        <View
            style={{
                paddingBottom: Math.max(0, insets.bottom - 10),
                height: "100%",
                display: "flex",
                flexDirection: "column",
                backgroundColor: colors.BACKGROUND
            }}
        >
            <KeyboardAvoidingView
                behavior="padding"
                style={{
                    display: "flex",
                    flexDirection: "column",
                    flexGrow: 1,
                    flexShrink: 1
                }}
            >
                <BackHeader title={user_pool[userID].firstName}/>
                <MessagesView messages={messages} userID={userID} />
                <InputBar
                    sendMessage={sendMessage}
                />
            </KeyboardAvoidingView>
        </View>
    );
};


const BackHeader = ({title}) => {
    const colors = useColors();
    const navigation = useNavigation();
    const insets = usePaddedSafeAreaInsets();

    return (
        <View style={{
            display: "flex",
            flexDirection: "row",
            paddingLeft: insets.left,
            paddingRight: insets.right,
            paddingTop: insets.top,
            paddingBottom: style_values.screen_padding.vertical,
            borderBottomColor: colors.SECONDARY_TEXT,
            borderBottomWidth: StyleSheet.hairlineWidth,
        }}>
            <TouchableOpacity
                style={{
                    display: "flex",
                    flexDirection: "row",
                    alignItems: "center"
                }}
                onPress={e => navigation.goBack()}
            >
                <Icon
                    height={28}
                    color={colors.TEXT}
                >chevron-left</Icon>
                <StyledText style={{fontSize: 28, fontWeight: 500, marginLeft: 29*.35}}>{title}</StyledText>
            </TouchableOpacity>
        </View>
    );
}

/**
 * @param {object} props
 * @param {import("../scripts/hardcoded_data").Message} props.message
 * @param {boolean} props.lastInMessageGroup
 */
const OtherBubble = ({message, lastInMessageGroup, ...props}) => {
    const colors = useColors();

    return (
        <View style={{
            display: "flex",
            flexDirection: "row",
            alignItems: "flex-end"
        }}>
            {
                lastInMessageGroup ?
                <Image
                    style={{
                        width: 40,
                        height: 40,
                        borderRadius: 20,
                        marginRight: 8,
                        marginVertical: -4
                    }}
                    source={{
                        uri: user_pool[message.senderId].profileImage
                    }}
                /> :
                <View
                    style={{
                        width: 40,
                        height: 40,
                        marginRight: 8,
                        marginVertical: -4
                    }}
                />
            }
            <View
                style={{
                    borderRadius: 16,
                    backgroundColor: Color(colors.FLAT1).mix(Color(colors.BACKGROUND), 0.8).hex(),
                    paddingHorizontal: 15,
                    paddingVertical: 5,
                    overflow: "hidden",
                    flexShrink: 1
                }}
            >
                <StyledText style={{fontSize: 15, lineHeight: 22}}>{message.content}</StyledText>
            </View>
            <View style={{width: 80}}/>
        </View>
    );
};

/**
 * @param {object} props
 * @param {import("../scripts/hardcoded_data").Message} props.message
 * @param {boolean} props.lastInMessageGroup
 */
const SelfBubble = ({message, lastInMessageGroup, ...props}) => {
    const colors = useColors();

    return (
        <View style={{
            display: "flex",
            flexDirection: "row-reverse",
            alignItems: "flex-end"
        }}>
            {
                lastInMessageGroup ?
                <Image
                    style={{
                        width: 40,
                        height: 40,
                        borderRadius: 20,
                        marginLeft: 8,
                        marginVertical: -4
                    }}
                    source={{
                        uri: selfUser.profileImage
                    }}
                /> :
                <View
                    style={{
                        width: 40,
                        height: 40,
                        marginLeft: 8,
                        marginVertical: -4
                    }}
                />
            }
            <View
                style={{
                    borderRadius: 16,
                    backgroundColor: Color(colors.FLAT1).mix(Color(colors.BACKGROUND), 0.8).hex(),
                    paddingHorizontal: 15,
                    paddingVertical: 5,
                    overflow: "hidden",
                    flexShrink: 1
                }}
            >
                <LinearGradient
                    style={{
                        position: "absolute",
                        top: 0,
                        left: 0,
                        bottom: 0,
                        right: 0
                    }}
                    colors={[colors.FLAT1, colors.FLAT2]}
                    start={{ x: 1.3, y: 0.5 }}
                    end={{ x: 0, y: 0.5 }}
                />
                <StyledText style={{color: colors.INVERTED_TEXT, fontSize: 15, lineHeight: 22}}>{message.content}</StyledText>
            </View>
            <View style={{width: 80}}/>
        </View>
    );
};

/**
 * @param {Object} props
 * @param {Array<import("../scripts/hardcoded_data").Message>} props.messages
 */
const MessagesView = ({messages, userID}) => {
    const insets = usePaddedSafeAreaInsets();

    return (
        <FlatList
            style={{
                paddingLeft: insets.left,
                paddingRight: insets.right,
                flexShrink: 1
            }}
            data={messages}
            contentContainerStyle={{
                flexGrow: 1,
                paddingTop: style_values.screen_padding.vertical,
                paddingBottom: style_values.screen_padding.vertical,
            }}
            ListEmptyComponent={
                <View
                    style={{
                        justifyContent: "center",
                        alignItems: "center",
                        flexGrow: 1
                    }}
                >
                    <StyledText
                        type="secondary"
                        style={{ textAlign: "center" }}
                    >This is the start of your conversation with {user_pool[userID].firstName}.</StyledText>
                </View>
            }
            keyExtractor={(item, index) => index}
            renderItem={({item: message, index, separators}) => {
                const lastInMessageGroup = messages[index + 1]?.senderId !== message.senderId;
                const BubbleComponent = selfUser.id == message.senderId ? SelfBubble : OtherBubble;

                return <>
                    <BubbleComponent message={message} lastInMessageGroup={lastInMessageGroup}/>
                    <View style={{height: lastInMessageGroup ? 10 : 3}}/>
                </>;
            }}
        />
    );
}

/**
 * @param {object} props
 * @param {(message: import("../scripts/hardcoded_data").Message) => void} props.sendMessage
 */
const InputBar = ({sendMessage, ...props}) => {
    const insets = usePaddedSafeAreaInsets();
    const colors = useColors();

    const [messageContent, setMessageContent] = useState("");

    return (
        <View
            style={{
                display: "flex",
                flexDirection: "row",
                alignItems: "flex-end",
                paddingVertical: 10,
                paddingLeft: insets.left,
                paddingRight: insets.right,
                borderTopColor: colors.SECONDARY_TEXT,
                borderTopWidth: StyleSheet.hairlineWidth
            }}
        >
            <Icon height={25} color={colors.SECONDARY_TEXT} style={{marginRight: 10, marginBottom: 4.5}}>fa:location-arrow</Icon>
            <Icon height={25} color={colors.SECONDARY_TEXT} style={{marginRight: 10, marginBottom: 4.5}}>bi:camera-fill</Icon>
            <StyledTextInput
                containerStyle={{
                    flexShrink: 1,
                    marginRight: 10,
                    lineHeight: 22
                }}
                value={messageContent}
                onChangeText={setMessageContent}
                placeholder="Enter a message"
                blurOnSubmit={false}
                clearButtonMode={"never"}
                multiline
            />
            {
                messageContent.length ?
                <TouchableOpacity
                    style={{
                        height: 25,
                        width: 25,
                        marginRight: 10
                    }}
                    onPress={e => {
                        sendMessage({
                            senderId: selfUser.id,
                            type: 1,
                            content: messageContent
                        });
                        setMessageContent("");
                    }}
                >
                    <MaskedView
                        style={{
                            height: "100%",
                            width: "100%"
                        }}
                        maskElement={
                            <Icon height={25} width={25} color={colors.SECONDARY_TEXT} style={{marginRight: 10, marginBottom: 4.5}}>fluent:send-filled</Icon>
                        }
                    >
                        <LinearGradient
                            style={{height: "100%", width: "100%"}}
                            colors={[colors.FLAT1, colors.FLAT2]}
                            start={{ x: 1.3, y: 0.5 }}
                            end={{ x: 0, y: 0.5 }}
                        />
                    </MaskedView>
                </TouchableOpacity> :
                <Icon height={25} width={25} color={colors.SECONDARY_TEXT} style={{marginRight: 10, marginBottom: 4.5}}>carbon:microphone-filled</Icon>
            }
        </View>
    );
};


export default MessagingScreen;