import { useNavigation } from "@react-navigation/native";
import React, { useEffect, useRef, useState, useMemo } from "react";
import { View, Image, Animated, Dimensions, Pressable, Easing } from "react-native";
import Svg, { Circle, Defs, LinearGradient, Path, RadialGradient, Stop } from "react-native-svg";
import { HeaderBar } from "../components/HeaderBar";
import Icon from "../components/Icon";
import StyledText from "../components/StyledText";
import distanceBetweenPoints from "../scripts/distanceBetweenPoints";
import getSelfLocation from "../scripts/getSelfLocation";
import { approximateLocations, matchScores, user_pool, __PROTOTYPE__TASK2_USER_ID } from "../scripts/hardcoded_data";
import seeded_rand from "../scripts/seeded_rand";
import { style_values } from "../scripts/styles";
import useColors from "../scripts/useColors";
import GlobalEvents from "../scripts/GlobalEvents";
import useTriggerRerender from "../scripts/useTriggerRerender";
import { selfUser } from "../scripts/hardcoded_data";

import usePaddedSafeAreaInsets from "../scripts/usePaddedSafeAreaInsets";
import { LinearGradient as ExpoLinearGradient } from "expo-linear-gradient";
import MaskedView from "@react-native-masked-view/masked-view";

const nearYouIDs = [29,30,31,32,33,34,35,36,37];

/**
 * The maximum size of bubbles. A bubble with a 100% match score would be the maximum size, but any
 * match score smaller will scale the bubble size down appropriately.
 * @constant
 * @type {Number}
 * @default 150
 */
const BUBBLE_MAX_SIZE = 150;

/**
 * The angle so start bubble outlines at in radians. 0 corresponds to the rightmost edge of the
 * bubble. Positive values go in the clockwise direction. Must be in the range [0, 2 * Math.PI).
 * @constant
 * @type {Number}
 * @default 0
 */
const BUBBLE_OUTLINE_START_ANGLE = 0;

/**
 * The width of the bubble outline.
 * @constant
 * @type {Number}
 * @default 3
 */
const BUBBLE_OUTLINE_WIDTH = 3;

/**
 * The width of the padding in between the bubble and the bubble outline.
 * @constant
 * @type {Number}
 * @default 3.5
 */
const BUBBLE_OUTLINE_PADDING = 3.5;

/**
 * The color to use for the outline. Can be a color string directly, or a callback function that
 * returns a color string. If a callback function, the callback will receive a ColorSpecification
 * (as returned by `useColor()`) as its argument.
 * @constant
 * @type {import("react-native").ColorValue|(colors: import("../scripts/useColors").ColorSpecification) => import("react-native").ColorValue}
 * @default (colors) => colors.FLAT1
 */
const BUBBLE_OUTLINE_COLOR = (colors) => colors.FLAT1;

/**
 * The font size to use for the match score labels around the bubbles.
 * @constant
 * @type {Number}
 * @default 12
 */
const BUBBLE_LABEL_FONT_SIZE = 13;

/**
 * The force of repulsion each bubble exerts on each other. The walls of the screen also exert a
 * repulsive force that prevents bubble from hitting up against the walls as often.
 * @constant
 * @type {Number}
 * @default 2
 */
const BUBBLE_MOTION_REPULSION = 2;

/**
 * The number of iterations each bubble motion update should go through. A greater number will will
 * make it more likely bubbles won't overlap, but will also take longer to calculate.
 * @constant
 * @type {Number}
 * @default 10
 */
const BUBBLE_MOTION_ITERATIONS = 10;

/**
 * The maximum distance (squared) a bubble can travel each iteration. This prevent bubbles from
 * exploding all over the place. It acts as a cap on the speed of bubbles.
 * @constant
 * @type {Number}
 * @default 4
 */
const MAX_DISPLACEMENT_SQUARED = 2;

/**
 * The amount of motion that a resting bubble has. A bubble that's going too slow will have its
 * motion scaled up according to this value to ensure a bubble is never just sitting there with 0
 * motion (unless this is set 0).
 * @constant
 * @type {Number}
 * @default 0.5
 */
const BUBBLE_MOTION_RESTING_JITTER = 0.5;

/**
 * The number of times to update the bubbles paths every second. This will scale up other values as
 * well, so increasing this value may require you to decrease other values, specifically
 * `BUBBLE_MOTION_RESTING_JITTER` to prevent bubbles from going too fast. The higher the number
 * here, the smoother the animation, but the more calculations need to happen every second.
 * @constant
 * @type {Number}
 * @default 20
 */
const BUBBLE_MOTION_UPDATE_RATE = 20;

/**
 * Controls how the radius of a bubble scales up with the match score. A small exponent makes
 * bubbles shrink slowly as match score goes down, and then more rapidly towards the end. A bigger
 * exponent exaggerates differences in match score more.
 * @constant
 * @type {(t: Number) => Number}
 * @default {(t) => t}
 */
const BUBBLE_SIZE = (t) => t ** 1;

const NearYouScreen = () => {
    const insets = usePaddedSafeAreaInsets();
    const colors = useColors();

    const triggerRerender = useTriggerRerender();

    // Whenever we toggle status, rerender and remove
    GlobalEvents.useListener("statusupdate", () => {
        triggerRerender();
    }, []);


    return (
        <View
            style={{
                display: "flex",
                flexDirection: "column",
                height: "100%"
            }}
        >
            <HeaderBar/>
            <View
                style={{
                    backgroundColor: colors.BACKGROUND,
                    flexGrow: 1
                }}
            >
                <CollisionNotifier userID={(selfUser.isActive) ? __PROTOTYPE__TASK2_USER_ID : null}/>
                <StyledText
                    type="h2"
                    style={{
                        paddingLeft: insets.left,
                        paddingRight: insets.right,
                        paddingTop: style_values.screen_padding.vertical,
                        paddingBottom: 5
                    }}
                >Others Near You</StyledText>
                <NearYouBody/>
            </View>
        </View>
    );
};

/**
 * Take a set of positions and updates their coordinates according to different forces on each
 * position. Each position has an X and Y coordinate as well as a radius for big the bubble at that
 * position is. Each bubble exerts a repulsive on one another, and the walls exert their own
 * repulsive force on the bubbles. The bubbles will move around according to the parameter constants
 * at the top of this file. The provided `positions` object has its data mutated directly.
 * @param {{ [key: import("../scripts/hardcoded_data").UserID]: {
 *      x: Number,
 *      y: Number,
 *      speedX: Number,
 *      speedY: Number,
 *      id: import("../scripts/hardcoded_data").UserID,
 *      radius: Number
 * } }} positions The old positions. This value will be mutated directly.
 * @param {Number} bodyHeight The maximum height the bubbles should stay in.
 * @param {Number} bodyWidth The maximum width the bubbles should stay in.
 */
const updateBubblePositions = (positions, bodyHeight, bodyWidth) => {
    // First make sure each node is within the boundaries.
    for (const ID in positions) {
        // Take care of the left side.
        if (positions[ID].y <= positions[ID].radius) {
            // Set a minimum y value so that the bubble cannot appear off the screen.
            positions[ID].y = positions[ID].radius;
            // Also make sure its speed is reset so that it doesn't think it's still going in the
            // direction off the screen.
            positions[ID].speedY = Math.max(0, positions[ID].speedY);
        }
        // Now do the right side.
        if (positions[ID].y >= bodyHeight - positions[ID].radius) {
            positions[ID].y = bodyHeight - positions[ID].radius;
            positions[ID].speedY = Math.min(0, positions[ID].speedY);
        }
        // Now do the top.
        if (positions[ID].x <= positions[ID].radius) {
            positions[ID].x = positions[ID].radius;
            positions[ID].speedX = Math.max(0, positions[ID].speedX);
        }
        // And the bottom.
        if (positions[ID].x >= bodyWidth - positions[ID].radius) {
            positions[ID].x = bodyWidth - positions[ID].radius;
            positions[ID].speedX = Math.min(0, positions[ID].speedX);
        }
    }

    /** @type {Array<import("../scripts/hardcoded_data").UserID>} */
    const IDs = Array.from(Object.keys(positions));

    // Do a few iterations so we're less likely to end up on a bad position.
    for (let iteration = 0; iteration < BUBBLE_MOTION_ITERATIONS; iteration++) {
        // Calculate repulsion on each node.
        for (let i1 = 0; i1 < IDs.length; i1++) {
            const node1 = positions[IDs[i1]];

            // First do a repulsion from each wall. This force is multiplied by 2 at the end whereas
            // the repulsive force between bubbles is not because the wall does not move away as
            // another bubble would, so it forces bubbles away from the wall at about the same
            // distance as bubbles do to other bubbles.
            // Do the left wall first.
            node1.speedX += BUBBLE_MOTION_REPULSION / (Math.max(node1.x - node1.radius, 1) ** 2) * 2;
            // Then right.
            node1.speedX -= BUBBLE_MOTION_REPULSION / (Math.max(bodyWidth - node1.x - node1.radius, 1) ** 2) * 2;
            // Then top.
            node1.speedY += BUBBLE_MOTION_REPULSION / (Math.max(node1.y - node1.radius, 1) ** 2) * 2;
            // Finally bottom.
            node1.speedY -= BUBBLE_MOTION_REPULSION / (Math.max(bodyHeight - node1.y - node1.radius, 1) ** 2) * 2;

            // Now do repulsion from every other node.
            for (let i2 = i1 + 1; i2 < IDs.length; i2++) {
                const node2 = positions[IDs[i2]];
                // Calculate the difference in distance.
                const dx = node2.x - node1.x;
                const dy = node2.y - node1.y;

                // If the distance between these two nodes is 0, perturb them so that they don't
                // perfectly overlap. It's a while loop in the extremely unlikely case that both
                // seeded_rands somehow return 0, then it'll try again until `dx` and `dy` aren't 0.
                let iterations = 0;
                while (dx == 0 && dy == 0) {
                    dx = seeded_rand(i1, i2, iterations, "dx") * (node1.radius + node2.radius) / 2 * 0.2;
                    dy = seeded_rand(i1, i2, iterations, "dy") * (node1.radius + node2.radius) / 2 * 0.2;
                    iterations++;
                }

                // This is the actual distance between the centers of the two bubbles.
                const actualDistance = Math.hypot(dx, dy);
                // Subtract off the bubble radii to use the distance between bubbles instead. This
                // makes bubbles repel from each other's edges instead of from each other's centers.
                const distance = Math.max(actualDistance - node1.radius - node2.radius, 1);
                // Calculate a total force from this distance.
                const force = BUBBLE_MOTION_REPULSION / (distance * distance);
                // Apply a force in the X and Y direction on each bubble.
                const fx = force * dx / distance;
                const fy = force * dy / distance;
                node1.speedX -= fx;
                node1.speedY -= fy;
                node2.speedX += fx;
                node2.speedY += fy;
            }

            // Update the positions based on speed.
            // Find the distance in each direction we need to move.
            let dx = node1.speedX / BUBBLE_MOTION_ITERATIONS;
            let dy = node1.speedY / BUBBLE_MOTION_ITERATIONS;
            const displacementSquared = dx * dx + dy * dy;
            if (displacementSquared > MAX_DISPLACEMENT_SQUARED) {
                const s = Math.sqrt(MAX_DISPLACEMENT_SQUARED / displacementSquared);
                dx = dx * s;
                dy = dy * s;
            }
            // Update the X and Y positionsr.
            node1.x += dx;
            node1.y += dy;
        }
    }

    // Normalize the motion for the next iteration. This keeps bubbles always moving. A bubble with
    // very little speed might come to a rest, but this prevents that from happening so that the
    // bubbles keep jittering around slowly.
    for (const ID in positions) {
        // If a bubble somehow has exactly 0 speed (which is very unlikely, but could technically
        // happen if the bubble was surrounded in just the right circumstances), give the bubble a
        // random speed so that it can't settle at 0 speed.
        if (positions[ID].speedX == 0 && positions[ID].speedY == 0) {
            const angle = seeded_rand(ID, "random speed angle") * Math.PI * 2;
            // We don't need to calculate a distance since the speed's distance is going to be
            // normalized anyway.
            positions[ID].speedX = Math.cos(angle);
            positions[ID].speedY = Math.sin(angle)
        }

        // Amplify the speed if we are going below `BUBBLE_MOTION_RESTING_JITTER`. If we are going
        // *faster* than that, we don't dampen the speed because we want bubbles to keep their
        // velocities and slow down naturally.
        const speed = Math.min(Math.hypot(positions[ID].speedX, positions[ID].speedY), 1);
        positions[ID].speedX *= BUBBLE_MOTION_RESTING_JITTER / speed;
        positions[ID].speedY *= BUBBLE_MOTION_RESTING_JITTER / speed;
    }
};

/**
 * This is the part of the screen that houses the moving bubbles.
 * @param {object} props
 */
const NearYouBody = (props) => {
    const bubbles = new Array(nearYouIDs.length);

    /** @type {import("react").MutableRefObject<{
     *      [key: import("../scripts/hardcoded_data").UserID]: {
     *          x: Number,
     *          y: Number,
     *          speedX: Number,
     *          speedY: Number,
     *          id: import("../scripts/hardcoded_data").UserID,
     *          radius: Number,
     *          animatedValue: Animated.AnimatedValueXY
     *      }
     * }>} */
    const positions = useRef({});

    // Keep track of the height and width of this portion of the screen so we can use it to
    // calculate where teh walls are, since we want the bubbles to stay inside the walls.
    const [bodyWidth, setBodyWidth] = useState(Dimensions.get("window").width);
    const [bodyHeight, setBodyHeight] = useState(Dimensions.get("window").height / 2);

    // Updates the positions ref, but doesn't actually return that value, so the return value is
    // discarded.
    void useMemo(() => {
        const missingIDs = new Set(nearYouIDs.map(ID => ID + ""));
        // Remove positions that are no longer on the screen.
        for (const ID in positions.current) {
            if (missingIDs.delete(ID + "")) continue;
            delete positions.current[ID];
        }
        // Add an initial position for IDs that are not yet in `positions`.
        for (const ID of missingIDs) {
            const radius = BUBBLE_SIZE(matchScores[ID]) * BUBBLE_MAX_SIZE / 2 + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH;
            const pos = {
                x: seeded_rand(ID, "width") * (bodyWidth - radius * 2),
                y: seeded_rand(ID, "height") * (bodyHeight - radius * 2)
            };
            const animatedValue = new Animated.ValueXY(pos);
            positions.current[ID] = {
                x: pos.x,
                y: pos.y,
                speedX: seeded_rand(ID, "speedX") * BUBBLE_MOTION_RESTING_JITTER,
                speedY: seeded_rand(ID, "speedY") * BUBBLE_MOTION_RESTING_JITTER,
                id: ID,
                radius: radius,
                animatedValue: animatedValue
            };
        }
        // `positions` should now have the same IDs as `nearYouIDs`.
    }, [nearYouIDs]);

    useEffect(() => {
        // Begin animating bubble positions.

        // This function calls `updateBubblePositions` to update the coordinates of each bubble, and
        // then uses those coordinates to animate the positions of bubbles.
        const doUpdate = () => {
            // Update X and Y coordinates of `positions`.
            updateBubblePositions(
                positions.current,
                bodyHeight,
                bodyWidth
            );

            // Use those coordinates to update positions of bubbles on screen.
            for (const ID in positions.current) {
                const node = positions.current[ID];
                Animated.timing(
                    node.animatedValue,
                    {
                        // `toValue` subtracts off the radius of the bubble so that the bubble's
                        // position corresponds to the center of the bubble instead of the top left.
                        toValue: { x: node.x - node.radius, y: node.y - node.radius },
                        // The duration is set to twice the frame rate. This means that the bubble
                        // is only halfway to its new position before the next position is
                        // calculated. This somewhat helps with preventing choppiness between
                        // frames.
                        duration: 1000 / BUBBLE_MOTION_UPDATE_RATE * 2,
                        useNativeDriver: true
                    }
                ).start();
            }
        }

        // Start doing updates.
        doUpdate();
        // Schedule an interval to perform consecutive updates.
        const interval = setInterval(doUpdate, 1000 / BUBBLE_MOTION_UPDATE_RATE);

        // On unmount, return a function to clear the interval.
        return () => clearInterval(interval);
    }, [nearYouIDs, bodyHeight, bodyWidth]);

    for (let i = 0; i < nearYouIDs.length; i++) {
        // Generate a bubble component for each bubble data point.
        bubbles[i] = (
            <Bubble
                key={nearYouIDs[i]}
                userID={nearYouIDs[i]}
                position={positions.current[nearYouIDs[i]]?.animatedValue ?? {x: 0, y: 0}}
            />
        );
    }

    // Stick all the bubbles into a parent <View>.
    return (
        <View
            style={{
                flexGrow: 1,
            }}
            onLayout={({nativeEvent: {layout: {height, width}}}) => {
                setBodyHeight(height);
                setBodyWidth(width);
            }}
        >
            {bubbles}
        </View>
    );
}

/**
 * Represents a single bubble on screen.
 * @param {object} props
 * @param {import("../scripts/hardcoded_data").UserID} props.userID
 * @param {{ x: (Number|Animated.AnimatedValue), y: Number|Animated.AnimatedValue }} props.position
 */
const Bubble = ({userID, position, ...props}) => {
    const colors = useColors();

    const user = user_pool[userID];
    const matchScore = matchScores[userID];
    // Get the bubble radius size.
    const bubbleRadius = BUBBLE_SIZE(matchScore) * BUBBLE_MAX_SIZE / 2;
    // We already know where teh start of the border outline is from `BUBBLE_OUTLINE_START_ANGLE`,
    // but we need to calculate where the border ends by adding an angle proportional to the match
    // score.
    const endAngle = BUBBLE_OUTLINE_START_ANGLE + Math.PI * 2 * matchScore;
    // This is the total spacing that will go around the bubble's image.
    const totalBubblePadding = BUBBLE_OUTLINE_WIDTH + BUBBLE_OUTLINE_PADDING;

    // This is the text that will appear on the percentage label.
    const labelText = Math.floor(matchScore * 100) + "%";
    // Calculate how wide the label will be once we put it on screen. This is just an approximation
    // since we can't know the exact width without using onLayout, but this serves as a decent
    // approximation for making other pixel measurement calculations.
    const approximateLabelWidth = Array.from(labelText).map(char =>
        // These widths were gotten just by measuring out each character and approximating their
        // measured width. It tends to give around the right answer, off by a pixel or two, which
        // isn't a big deal for this use case.
        ({
            "0": 0.71 + 1/300,
            "1": 0.35,
            "2": 0.57 + 2/300,
            "3": 0.60,
            "4": 0.64 + 2/300,
            "5": 0.62 + 1/300,
            "6": 0.64,
            "7": 0.54,
            "8": 0.63 + 1/300,
            "9": 0.64,
            "%": 0.87 + 2/300
        })[char] || 0
    ).reduce((a, b) => a + b, 0) * BUBBLE_LABEL_FONT_SIZE + 8;

    // Since the text appears *after* the edge of the bubble outline (as opposed to being centered
    // right on top), we need to calculate a new angle to position the label. We do that by figuring
    // out its chord length along the edge of the circle, calculating the angle of that chord, and
    // then dividing that by two to correspond to the center of the label.
    const R = bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2;
    const D = approximateLabelWidth;
    const labelY = Math.sqrt(4 * D * D * R * R - D ** 4) / (2 * R);
    const labelX = Math.sqrt(R * R - labelY * labelY) * (Math.sqrt(2 * R * R) < D ? -1 : 1);
    const labelAdditionalRotation = Math.atan2(labelY, labelX) / 2;
    const labelRotationAngle = (Math.PI * 2 + BUBBLE_OUTLINE_START_ANGLE + labelAdditionalRotation + Math.PI / 2 + matchScore * Math.PI * 2) % (Math.PI * 2);

    // Get the color of the bubble outline.
    const outlineColor = typeof BUBBLE_OUTLINE_COLOR == "function" ? BUBBLE_OUTLINE_COLOR(colors) : BUBBLE_OUTLINE_COLOR;

    // Whether this label needs to be flipped. Some labels appear at the bottom of the bubble, but
    // that would make the text look upside down, so this determined whether we need to do an
    // additional flip at the end so that it appears right side up.
    const flipLabel = Math.PI * 0.5 < labelRotationAngle && labelRotationAngle < Math.PI * 1.5;

    const navigation = useNavigation();

    // If the match score ie big enough, the text label will begin to overlap with the border. A
    // match score of 99% for example will have a border filled up to 99%, and then the "99%" text
    // label appears next to that. Since there's only 1% left (i.e., 3.6 degrees) of blank space,
    // the text label will go beyond that, and will show up on top of the start of outline. In this
    // case, we add a small fade behind the text label that shows the background color. and covers
    // part of the outline so that the text label appears to cover up over it.
    let textOverlapFade = null;
    // First check if the text label will go past 360 degrees.
    if ((endAngle - BUBBLE_OUTLINE_START_ANGLE + labelAdditionalRotation * 2) % (Math.PI * 2) < Math.PI && (endAngle - BUBBLE_OUTLINE_START_ANGLE) > Math.PI) {
        // Calculate the overflow angle (i.e., the amount we need to hide).
        const angleToHide = (endAngle - BUBBLE_OUTLINE_START_ANGLE + labelAdditionalRotation * 2) % (Math.PI * 2);
        // Turn that angle into a chord length.
        const distanceToHide = Math.hypot(R * (Math.cos(angleToHide) - 1), R * Math.sin(angleToHide));

        // Set textOverlapFade to the SVG that will cover up part of the outline.
        textOverlapFade = <Svg
            pointerEvents="none"
            viewBox={`${-1 * (totalBubblePadding + bubbleRadius)} ${-1 * (totalBubblePadding + bubbleRadius)} ${2 * (totalBubblePadding + bubbleRadius)} ${2 * (totalBubblePadding + bubbleRadius)}`}
            style={{
                position: "absolute",
                top: 0,
                left: 0,
                height: 2 * (totalBubblePadding + bubbleRadius),
                width: 2 * (totalBubblePadding + bubbleRadius)
            }}
        >
            <Defs>
                {/* Instead of just inserting a solid color on top, we use a gradient so it the out-
                    line looks more like its fading away instead of just an abrupt stop. */}
                <LinearGradient
                    id="fadeGradient"
                    gradientUnits="userSpaceOnUse"
                    x1={Math.cos(BUBBLE_OUTLINE_START_ANGLE) * R}
                    y1={Math.sin(BUBBLE_OUTLINE_START_ANGLE) * R}
                    x2={Math.cos(BUBBLE_OUTLINE_START_ANGLE + angleToHide) * R}
                    y2={Math.sin(BUBBLE_OUTLINE_START_ANGLE + angleToHide) * R}
                >
                    <Stop
                        offset={Math.min(0, distanceToHide - BUBBLE_LABEL_FONT_SIZE / 2) / distanceToHide}
                        stopColor={colors.BACKGROUND}
                    />
                    <Stop
                        offset={(distanceToHide - BUBBLE_LABEL_FONT_SIZE / 2) / distanceToHide}
                        stopColor={colors.BACKGROUND}
                    />
                    <Stop
                        offset={1}
                        stopColor={outlineColor}
                    />
                </LinearGradient>
            </Defs>
            {/* This is the path that overlaps with the outline and covers it using a gradient. */}
            <Path
                strokeLinecap="round"
                strokeWidth={BUBBLE_OUTLINE_WIDTH}
                stroke="url(#fadeGradient)"
                fill="none"
                d={
                    `M${
                        Math.cos(BUBBLE_OUTLINE_START_ANGLE) * (bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2)
                    } ${
                        Math.sin(BUBBLE_OUTLINE_START_ANGLE) * (bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2)
                    }A${
                        bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2
                    } ${
                        bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2
                    } 0 0 1 ${
                        Math.cos(BUBBLE_OUTLINE_START_ANGLE + angleToHide) * (bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2)
                    } ${
                        Math.sin(BUBBLE_OUTLINE_START_ANGLE + angleToHide) * (bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2)
                    }`
                }
            />
            {/* This circle is added at the very tip of of the end of the outline. If the match
                score is 100%, the tip will lose the circle at the edge because it'll get covered
                up. This make sure it still looks round by adding the round tip back. */}
            <Circle
                cx={Math.cos(endAngle) * R}
                cy={Math.sin(endAngle) * R}
                r={BUBBLE_OUTLINE_WIDTH / 2}
                fill={outlineColor}
            />
        </Svg>
    }

    // The outline is the circle that goes around the image. It's an arc where the angle corresponds
    // to the match score percentage.
    const outline = (
        <Svg
            pointerEvents="none"
            viewBox={`${-1 * (totalBubblePadding + bubbleRadius)} ${-1 * (totalBubblePadding + bubbleRadius)} ${2 * (totalBubblePadding + bubbleRadius)} ${2 * (totalBubblePadding + bubbleRadius)}`}
            style={{
                position: "absolute",
                top: 0,
                left: 0,
                height: 2 * (totalBubblePadding + bubbleRadius),
                width: 2 * (totalBubblePadding + bubbleRadius)
            }}
        >
            {
                // If the match score is 1 (or *extremely* close to 1), we can draw a full circle
                // instead of an arc.
                matchScore > 0.999 ?
                <Circle
                    strokeWidth={BUBBLE_OUTLINE_WIDTH}
                    stroke={outlineColor}
                    fill="none"
                    r={bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2}
                    cx={0}
                    cy={0}
                /> :
                <Path
                    strokeLinecap="round"
                    strokeWidth={BUBBLE_OUTLINE_WIDTH}
                    stroke={outlineColor}
                    fill="none"
                    d={
                        `M${
                            Math.cos(BUBBLE_OUTLINE_START_ANGLE) * (bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2)
                        } ${
                            Math.sin(BUBBLE_OUTLINE_START_ANGLE) * (bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2)
                        }A${
                            bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2
                        } ${
                            bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2
                        } 0 ${
                            Math.round(matchScore)
                        } 1 ${
                            Math.cos(endAngle) * (bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2)
                        } ${
                            Math.sin(endAngle) * (bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2)
                        }`
                    }
                />
            }
        </Svg>
    );

    // A subtle box shadow is also added around the profile image to make it stand out a tiny bit
    // more. Since React Native kinda sucks with box shadows it's also implemented as an SVG with a
    // <radialGradient>.
    const shadow = (
        <Svg
            pointerEvents="none"
            viewBox={`${-1 * (totalBubblePadding + bubbleRadius)} ${-1 * (totalBubblePadding + bubbleRadius)} ${2 * (totalBubblePadding + bubbleRadius)} ${2 * (totalBubblePadding + bubbleRadius)}`}
            style={{
                position: "absolute",
                top: 0,
                left: 0,
                height: 2 * (totalBubblePadding + bubbleRadius),
                width: 2 * (totalBubblePadding + bubbleRadius)
            }}
        >
            <Defs>
                <RadialGradient
                    id="shadow"
                    cx={0}
                    cy={0}
                    fx={0}
                    fy={0}
                    r={bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2}
                    gradientUnits="userSpaceOnUse"
                >
                    <Stop offset={bubbleRadius / (bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2)} stopColor={outlineColor} stopOpacity={.3}/>
                    <Stop offset={(bubbleRadius + BUBBLE_OUTLINE_PADDING) / (bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2)} stopColor={outlineColor} stopOpacity={0}/>
                </RadialGradient>
            </Defs>
            <Circle
                stroke="none"
                fill="url(#shadow)"
                r={bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2}
                cx={0}
                cy={0}
            />
        </Svg>
    );

    // This is the profile image that appears in the circle.
    const profileImage = (
        <View
            style={{
                height: bubbleRadius * 2,
                width: bubbleRadius * 2,
                borderRadius: bubbleRadius,
                position: "absolute",
                top: totalBubblePadding,
                left: totalBubblePadding,
                overflow: "hidden"
            }}
        >
            <Pressable
                style={{
                    height: "100%",
                    width: "100%"
                }}
                onPress={({nativeEvent: {locationX, locationY}}) => {
                    // Since a Pressable doesn't really respect border radiuses, we need to check if
                    // the touch originated from a point within the circle. We do this by checking
                    // that its distance from the center is at most `bubbleRadius`.
                    if (Math.hypot(bubbleRadius - locationX, bubbleRadius - locationY) > bubbleRadius) {
                        return;
                    }
                    // When the user presses the image of a bubble, it should take them to the profile
                    // page for that person.

                    let distance = distanceBetweenPoints(
                        getSelfLocation().coords,
                        approximateLocations[userID].coords,
                        "mi"
                    );
                    distance = Math.max(0.1, distance >= 10 ? Math.round(distance) : Math.round(distance * 1e1) / 1e1) + " mi";
                    navigation.navigate("Other's Profile Screen", { userID, distance });
                }}
            >
                <Image
                    style={{
                        height: "100%",
                        width: "100%"
                    }}
                    source={{
                        uri: user.profileImage
                    }}
                />
            </Pressable>
        </View>
    );

    // Finally, we draw the label. Its angle and position are determined by the match score so that
    // it lines up with the edge of the outline.
    const label = (
        <View
            pointerEvents="none"
            style={{
                height: "100%",
                width: "100%",
                position: "absolute",
                left: 0,
                top: 0,
                justifyContent: "center",
                alignItems: "center",
            }}
        >
            <View
                style={{
                    transform: [
                        { rotateZ: labelRotationAngle + "rad" },
                        { translateY: -(bubbleRadius + BUBBLE_OUTLINE_PADDING + BUBBLE_OUTLINE_WIDTH / 2) },
                        { rotateZ: flipLabel ? "180deg" : "0deg" }
                    ],
                    paddingHorizontal: 4,
                }}
            >
                <StyledText
                    style={{
                        fontWeight: "bold",
                        color: colors.TEXT,
                        textAlign: "center",
                        fontSize: BUBBLE_LABEL_FONT_SIZE
                    }}
                >{labelText}</StyledText>
            </View>
        </View>
    );

    // Put all the components above into one unified bubble. It uses an Animated.View so that it can
    // have its X and Y translation animated.
    return (
        <Animated.View
            style={{
                position: "absolute",
                height: 2 * (bubbleRadius + totalBubblePadding),
                width: 2 * (bubbleRadius + totalBubblePadding),
                overflow: "visible",
                transform: [
                    { translateX: position.x },
                    { translateY: position.y }
                ]
            }}
        >
            {outline}
            {textOverlapFade}
            {shadow}
            {profileImage}
            {label}
        </Animated.View>
    );
};

/**
 * The button at the top of the screen that allows a user to access an ongoing collision.
 * @param {object} props
 * @param {import("../scripts/hardcoded_data").UserID|null} [props.userID] The ID of the user to
 *      show in the notifier.
 */
const CollisionNotifier = ({userID, ...props}) => {
    const user = user_pool[userID];
    const insets = usePaddedSafeAreaInsets();
    const colors = useColors();
    const navigation = useNavigation();
    const [isBeingPressed, setIsBeingPressed] = useState(false);
    const [totalWidth, setTotalWidth] = useState(Dimensions.get("window").width - style_values.screen_padding.horizontal * 2);

    const GRADIENT_LOOP_TIME = 30000;
    const NUMBER_OF_GRADIENT_SCREENS = 4;

    const gradientTranslation = useRef(new Animated.Value(-NUMBER_OF_GRADIENT_SCREENS)).current;
    useEffect(() => {
        const animation = Animated.loop(
            Animated.timing(
                gradientTranslation,
                {
                    toValue: 0,
                    useNativeDriver: true,
                    duration: GRADIENT_LOOP_TIME,
                    easing: Easing.linear,
                    isInteraction: false
                }
            )
        );
        animation.start();
        return () => void animation.stop();
    }, []);

    const AnimatedExpoLinearGradient = useMemo(() => Animated.createAnimatedComponent(ExpoLinearGradient), []);

    if (!user) return null;

    return (
        <View
            style={{
                paddingLeft: insets.left,
                paddingRight: insets.right,
                paddingTop: style_values.screen_padding.vertical
            }}
        >
            <Pressable
                style={{
                    borderRadius: style_values.border_radius,
                    paddingHorizontal: 10,
                    paddingVertical: 5,
                    display: "flex",
                    flexDirection: "row",
                    alignItems: "center"
                }}
                onPressIn={e => setIsBeingPressed(true)}
                onPressOut={e => setIsBeingPressed(false)}
                onPress={e => {
                    navigation.navigate("Collision Screen", { userID });
                }}
            >
                <MaskedView
                    onLayout={({nativeEvent: {layout: {width}}}) => {
                        setTotalWidth(width);
                    }}
                    style={{
                        position: "absolute",
                        top: 0,
                        left: 0,
                        bottom: 0,
                        right: 0
                    }}
                    maskElement={
                        <View
                            style={{
                                height: "100%",
                                width: "100%",
                                borderColor: "black",
                                borderWidth: 2.5,
                                borderRadius: style_values.border_radius,
                                paddingHorizontal: 7.5,
                                paddingVertical: 2.5,
                                display: "flex",
                                flexDirection: "row",
                                alignItems: "center",
                                backgroundColor: isBeingPressed ? "black" : "transparent",
                            }}
                        >
                            <View
                                style={{
                                    aspectRatio: 1/1,
                                    display: "flex",
                                    justifyContent: "center",
                                    alignItems: "center",
                                    alignSelf: "stretch"
                                }}
                            >
                                <View
                                    style={{
                                        aspectRatio: 1/1,
                                        height: 40,
                                        borderRadius: 40,
                                        backgroundColor: "black"
                                    }}
                                />
                            </View>
                            <View
                                style={{
                                    flexGrow: 1,
                                    flexShrink: 1,
                                    marginHorizontal: 10
                                }}
                            />
                            <Icon
                                height={24}
                                color="black"
                            >chevron-right</Icon>
                        </View>
                    }
                >
                    <AnimatedExpoLinearGradient
                        style={{
                            position: "absolute",
                            top: 0,
                            left: 0,
                            bottom: 0,
                            width: (NUMBER_OF_GRADIENT_SCREENS + 1) * 100 + "%",
                            transform: [
                                { translateX: Animated.multiply(gradientTranslation, totalWidth) }
                            ]
                        }}
                        colors={[colors.FLAT1, colors.FLAT1, colors.FLAT2, colors.FLAT2, colors.FLAT4, colors.FLAT4, colors.FLAT3, colors.FLAT3, colors.FLAT1, colors.FLAT1, colors.FLAT2, colors.FLAT2]}
                        locations={[0/6/5, 1/6/5, 5/6/5, 7/6/5, 11/6/5, 13/6/5, 17/6/5, 19/6/5, 23/6/5, 25/6/5, 29/6/5, 30/6/5]}
                        start={{ x: 0, y: 0 }}
                        end={{ x: 1, y: 1 }}
                    />
                </MaskedView>
                <View
                    style={{
                        aspectRatio: 1/1,
                        display: "flex",
                        justifyContent: "center",
                        alignItems: "center",
                        alignSelf: "stretch"
                    }}
                >
                    <View
                        style={{
                            aspectRatio: 1/1,
                            height: 40,
                            borderRadius: 40,
                            overflow: "hidden",
                            justifyContent: "center",
                            alignItems: "center",
                            backgroundColor: isBeingPressed ? colors.INVERTED_TEXT : "transparent"
                        }}
                    >
                        <Image
                            style={{
                                aspectRatio: 1/1,
                                height: 35,
                                borderRadius: 30
                            }}
                            source={{
                                uri: user.profileImage
                            }}
                        />
                    </View>
                </View>
                <View
                    style={{
                        display: "flex",
                        flexDirection: "column",
                        flexGrow: 1,
                        flexShrink: 1,
                        marginHorizontal: 10,
                        alignItems: "stretch"
                    }}
                >
                    <StyledText style={{
                        fontSize: 14,
                        color: isBeingPressed ? colors.INVERTED_TEXT : colors.TEXT,
                        fontWeight: 500
                    }}>View collision with</StyledText>
                    <View
                        style={{
                            display: "flex",
                            flexDirection: "row",
                            alignItems: "baseline",
                            width: "100%"
                        }}
                    >
                        <StyledText
                            numberOfLines={1}
                            style={{
                                fontSize: 20,
                                fontWeight: 700,
                                flexShrink: 1,
                                color: isBeingPressed ? colors.INVERTED_TEXT : colors.TEXT
                            }}
                        >{user.firstName} {user.lastName} </StyledText>
                        <StyledText
                            type="secondary"
                            style={{
                                fontSize: 14,
                                color: isBeingPressed ? colors.INVERTED_TEXT : colors.SECONDARY_TEXT
                            }}
                        >({Math.floor(matchScores[userID] * 100)}%)</StyledText>
                    </View>
                </View>
                <Icon
                    height={24}
                    color={colors.INVERTED_TEXT}
                    style={{
                        opacity: isBeingPressed ? 1 : 0
                    }}
                >chevron-right</Icon>
            </Pressable>
        </View>
    );
};

export default NearYouScreen;