import React, { useMemo, useState, useRef, useCallback, useEffect } from "react";
import { Animated, FlatList, Image, View, Pressable, Dimensions, Easing, TouchableOpacity, TextInput, KeyboardAvoidingView } from "react-native";
import { HeaderBar } from "../components/HeaderBar";
import StyledText from "../components/StyledText";
import { style_values } from "../scripts/styles";
import useColors from "../scripts/useColors";
import usePaddedSafeAreaInsets from "../scripts/usePaddedSafeAreaInsets";
import { selfUser, user_pool, network_connections } from "../scripts/hardcoded_data";
import { LinearGradient } from "expo-linear-gradient";
import { PanGestureHandler, PinchGestureHandler, State } from "react-native-gesture-handler";
import Icon from "../components/Icon";
import { useNavigation } from "@react-navigation/native";
import seeded_rand from "../scripts/seeded_rand";
import SearchBar from "../components/SearchBar";
import Color from "color";
import GlobalEvents from "../scripts/GlobalEvents";

/**
 * The size of each bubble within the network, in screen pixels.
 * @constant
 * @type {Number}
 * @default 65
 */
const BUBBLE_SIZE = 65;

/**
 * The number of bubbles in the first concentric ring of bubbles in the network. This number will
 * determine the number of bubbles in outer rings as well using a similar density. Should be a
 * positive integer.
 * @constant
 * @type {Number}
 * @default 6
 */
const NETWORK_DENSITY = 6;

/**
 * The average radius of the first concentric ring of bubbles. It is the number of pixels between
 * the *center* of the centric self bubble, and the centers of each bubble in the first ring. This
 * means `BUBBLE_SIZE` is not taken into account, and should be included here (i.e. the number of
 * pixels *between* each ring's bubbles is this number, minus `BUBBLE_SIZE`).
 * @constant
 * @type {Number}
 * @default BUBBLE_SIZE * 1.7
 */
const NETWORK_RING_RADIUS = BUBBLE_SIZE * 1.7;

/**
 * The number of milliseconds it takes a bubble to move from one spot to another. This can happen
 * when the network changes and the bubbles' positions are recalculated.
 * @constant
 * @type {Number}
 * @default 300
 */
const BUBBLE_MOVEMENT_DURATION = 300;

/**
 * A function that determines how opaque each edge in the network should be given the edge's length.
 * @constant
 * @param {Number} length The length of the edge to return the opacity for.
 * @returns {Number} The opacity for this edge. The number should be in the range [0,1].
 */
const EDGE_OPACITY = (length) => Math.min(Math.max((NETWORK_RING_RADIUS * .8 / length) ** 2, 0.05), 0.75);

/**
 * The minimum zoom level allowed for the network (i.e., how much the user is allowed to zoom out).
 * It should be a number in the range (0, 1].
 * @constant
 * @type {Number}
 * @default 0.3
 */
const MIN_ZOOM = 0.3;

/**
 * The maximum zoom level allowed for the network (i.e., how much the user is allowed to zoom in).
 * It should be a number in the range [1,Infinity).
 * @constant
 * @type {Number}
 * @default 2.5
 */
const MAX_ZOOM = 2.5;

/**
 * The Networking screen gets shown when the user goes to the networking tab using the navbar at the
 * bottom of the screen.
 * @param {object} props
 */
const NetworkScreen = (props) => {
    const insets = usePaddedSafeAreaInsets();
    const colors = useColors();

    return (
        <View
            style={{
                display: "flex",
                flexDirection: "column",
                height: "100%"
            }}
        >
            <HeaderBar/>
            <View
                style={{
                    backgroundColor: colors.BACKGROUND,
                    paddingTop: style_values.screen_padding.vertical,
                    flexGrow: 1,
                    display: "flex",
                    flexDirection: "column"
                }}
            >
                <StyledText type="h2" style={{
                    paddingLeft: insets.left,
                    paddingRight: insets.right
                }}>Your Friend Network</StyledText>
                <NetworkBody/>
            </View>
        </View>
    );
};

/**
 * The main network area. This includes everything between the "Your Friend Network" heading and the
 * menu bar at the bottom of the screen.
 * @param {object} props
 */
const NetworkBody = (props) => {
    // // The number of milliseconds in the past the timeline is currently scrolled to.
    // const [timeDelta, setTimeDelta] = useState(0);

    const insets = usePaddedSafeAreaInsets();
    const colors = useColors();

    const [allConnections, setAllConnections] = useState(network_connections.slice());
    // Possibly do some filtering here if we choose to implement that
    const [hiddenIndices, setHiddenIndices] = useState([]);

    // Whenever we add or remove a connection, change `allConnections` to match.
    GlobalEvents.useListener("connectionchange", ({added, removed}) => {
        const newConnections = allConnections.slice();
        let numRemoved = 0;
        for (let i = newConnections.length - 1; i >= 0 && numRemoved < removed.size; i--) {
            if (removed.has(newConnections[i].id)) {
                newConnections.splice(i, 1);
                numRemoved++;
            }
        }

        for (const ID of added) {
            newConnections.push({
                id: ID,
                connectedDate: new Date()
            });
        }

        setAllConnections(newConnections);
    }, [allConnections]);

    // Uncomment this section to have nodes added and removed from the network one by one to test
    // how the network adapts to losing and receiving nodes.
    /*

    // Incrementally removes bubbles until the network is empty and then adds them back in one by
    // one and then repeat.
    const [decreasing, setDecreasing] = useState(true);
    useEffect(() => {
        setTimeout(() => {
            if (decreasing) {
                setHiddenIndices(hiddenIndices.concat([hiddenIndices.length]))
                if (hiddenIndices.length + 1 >= allConnections.length) setDecreasing(false);
            } else {
                setHiddenIndices(hiddenIndices.slice(0, hiddenIndices.length - 1));
                if (hiddenIndices.length == 0) setDecreasing(true);
            }
        }, 2000);
    }, [hiddenIndices]);

    */

    /** @type React.MutableRefObject<TextInput> */
    const searchRef = useRef();
    const [searchQuery, setSearchQuery] = useState("");
    const [selectedType, setSelectedType] = useState(() => GraphicalNetwork);
    const NetworkComponent = searchQuery.trim().length > 0 ? ListNetwork : selectedType;
    const [isPressingTypeSwitcher, setIsPressingTypeSwitcher] = useState(false);

    return (
        <KeyboardAvoidingView
            enabled={NetworkComponent == ListNetwork}
            behavior="padding"
            keyboardVerticalOffset={insets.top + 60}
            style={{
                display: "flex",
                flexDirection: "column",
                flexGrow: 1,
            }}
        >
            <View
                style={{
                    paddingLeft: insets.left,
                    paddingRight: insets.right,
                    paddingVertical: 5,
                    display: "flex",
                    flexDirection: "row",
                    alignItems: "center"
                }}
            >
                <SearchBar
                    ref={searchRef}
                    placeholder="Search Friends"
                    autoCapitalize="words"
                    autoCorrect={false}
                    value={searchQuery}
                    onChangeText={text => {
                        setSearchQuery(text);
                        const words = text.trim().toLowerCase().split(/\s+/);

                        const hiddenIndices = [];

                        for (let i = 0; i < allConnections.length; i++) {
                            const name = nameOfUser(user_pool[allConnections[i].id]).toLowerCase();
                            for (const word of words) {
                                if (!name.includes(word)) {
                                    hiddenIndices.push(i);
                                    break;
                                }
                            }
                        }

                        setHiddenIndices(hiddenIndices);
                    }}
                    onCancel={() => {
                        setSearchQuery("");
                        setHiddenIndices([]);
                        searchRef.current?.blur();
                    }}
                    containerStyle={{
                        flexShrink: 1,
                        flexGrow: 1,
                        marginRight: style_values.screen_padding.horizontal
                    }}
                />
                <Pressable
                    onPressIn={e => setIsPressingTypeSwitcher(true)}
                    onPressOut={e => setIsPressingTypeSwitcher(false)}
                    onPress={e => {
                        if (NetworkComponent == ListNetwork) {
                            setSearchQuery("");
                            searchRef.current.blur();
                            setHiddenIndices([]);
                            setSelectedType(() => GraphicalNetwork);
                        } else {
                            setSelectedType(() => ListNetwork)
                        }
                    }}
                >
                    <Icon width={24} height={24} color={isPressingTypeSwitcher ? colors.TEXT : colors.SECONDARY_TEXT}>
                        {
                            NetworkComponent == ListNetwork ?
                            "bxs:network-chart" :
                            "foundation:list"
                        }
                    </Icon>
                </Pressable>
            </View>
            <NetworkComponent
                searchBarRef={searchRef}
                allConnections={allConnections}
                hiddenIndices={hiddenIndices}
            />
            {/* <DraggableTimeline
                onChangeTimeDelta={setTimeDelta}
                timeDelta={timeDelta}
            /> */}
        </KeyboardAvoidingView>
    );
};

/**
 * Returns the full name of a user. By default, it's just "FirstName LastName" separated by a
 * single space, but it can be changed for example to be "LastName, FirstName" or another format.
 * @param {import("../scripts/hardcoded_data").User} user The user to get the name for.
 * @returns {String} The formatted name of the user.
 */
const nameOfUser = (user) => {
    return `${user.firstName} ${user.lastName}`;
}

/**
 * A list view of the network. This version of the network shows up when the user searches a name in
 * the search bar, or when they toggle to this view instead of the graphical view.
 * @param {object} props
 * @param {Array<import("../scripts/hardcoded_data").Connection>} props.allConnections
 * @param {Array<Number>} props.hiddenIndices
 * @param {React.MutableRefObject<TextInput>} props.searchBarRef
 */
const ListNetwork = ({allConnections, hiddenIndices, searchBarRef, ...props}) => {
    const colors = useColors();
    const insets = usePaddedSafeAreaInsets();
    const navigation = useNavigation();

    // Sorts allConnections by alphabetical order of name, but instead of sorting allConnections
    // itself, it sorts a list of indices, so that we can still use the indices in `hiddenIndices`.
    const sortedIndices = useMemo(() =>
        allConnections.map((_, i) => i).sort((index1, index2) =>
            nameOfUser(user_pool[allConnections[index1].id]).localeCompare(nameOfUser(user_pool[allConnections[index2].id]))
        )
    , [allConnections]);
    hiddenIndices = useMemo(() => new Set(hiddenIndices), [hiddenIndices]);

    const Row = useCallback(({user}) => {
        const colors = useColors();
        const [isPressing, setIsPressing] = useState(false);

        return <Pressable
            style={{
                display: "flex",
                flexDirection: "row",
                paddingLeft: insets.left + 3,
                paddingRight: insets.right + 3,
                paddingVertical: 8,
                height: 48,
                alignItems: "center",
                backgroundColor: isPressing ? Color(colors.BACKGROUND).mix(Color(colors.TEXT), .1).hex() : colors.BACKGROUND
            }}
            onPressIn={() => setIsPressing(true)}
            onPressOut={() => setIsPressing(false)}
            onPress={e => {
                navigation.navigate("Other's Profile Screen", {
                    userID: user.id
                });
            }}
        >
            <Image
                style={{
                    aspectRatio: 1/1,
                    height: 32,
                    borderRadius: 32,
                    marginRight: 10
                }}
                source={{
                    uri: user.profileImage
                }}
            />
            <StyledText style={{fontSize: 18, flexGrow: 1}}>{nameOfUser(user)}</StyledText>
            <Icon height={16} color={colors.TEXT}>fluent:ios-chevron-right-regular</Icon>
        </Pressable>
    }, []);

    const list = useMemo(() =>
        <FlatList
            data={sortedIndices.filter(index => !hiddenIndices.has(index))}
            keyExtractor={(connectionIndex) => allConnections[connectionIndex].id}
            renderItem={({item: connectionIndex}) => <Row user={user_pool[allConnections[connectionIndex].id]}/>}
            ItemSeparatorComponent={() =>
                <View style={{
                    backgroundColor: colors.SECONDARY_TEXT,
                    height: 1,
                    marginRight: insets.right,
                    marginLeft: insets.left
                }}/>
            }
            ListEmptyComponent={() =>
                <View
                    style={{
                        display: "flex",
                        flexDirection: "row",
                        paddingLeft: insets.left + 3,
                        paddingRight: insets.right + 3,
                        paddingVertical: 8,
                        alignItems: "center",
                        justifyContent: "center",
                        minHeight: 48
                    }}
                >
                    <StyledText type="secondary" style={{textAlign: "center"}}>Couldn't find any friends with that name :(</StyledText>
                </View>
            }
            style={{
                height: "100%"
            }}
            keyboardShouldPersistTaps={"handled"}
        />
    , [sortedIndices, hiddenIndices]);

    return (
        <View
            style={{
                flexGrow: 1,
                flexBasis: 0
            }}
        >
            <View style={{
                backgroundColor: colors.SECONDARY_TEXT,
                height: 1,
                marginRight: insets.right,
                marginLeft: insets.left
            }}/>
            {list}
        </View>
    );
};

/**
 * Adds a plural or singular version of units given a number.
 * @param {Number} number The number to add units to.
 * @param {String} singular The singular form of the units to use.
 * @param {String} plural The plural form of the units to use.
 * @returns {String} The number followed by the appropriate units.
 */
const unitify = (number, singular, plural) => {
    if (number == 1) return "1 " + singular;
    return number + " " + plural;
};

/**
 * This function assigns initial positions to every bubble in the network. These positions are never
 * actually seen by the user however because they only serve as initial positions for
 * `updatePositions`. In other words, the first time the network is mounted, this function assigns
 * positions to all the bubbles to act as a seed, and then immediately after, `updatePositions` is
 * called. Every time after that, `updatePositions` works on the previous network positions, so this
 * function is never called again after the initial mount.
 *
 * The bubbles are placed in concentric rings around the center self bubble, with the closer rings
 * being the connections that appear first in the list, and growing outwards from there.
 *
 * @param {Array<{id: import("../scripts/hardcoded_data").UserID}>} connections The list of
 *      connections to assign positions for. The only property needed is the user's ID to assign a
 *      position that will correspond to the same user every time.
 * @param {Iterable<Number>} hiddenIndices The list of indices for the `connections` array that
 *      indicates which connections should be hidden. Hidden connections are assigned a position of
 *      (0, 0).
 * @returns {Array<{x: Number, y: Number, id: import("../scripts/hardcoded_data").UserID}>} The list
 *      of positions for each connection. The order of each position will be in the same order as
 *      was provided for `connections`, but each object will still have an ID assigned to it so that
 *      that can be used instead.
 */
const assignInitialPositions = (connections, hiddenIndices) => {
    const positions = new Array(connections.length);
    connections = JSON.parse(JSON.stringify(connections));

    const hiddenIndicesSetClone = new Set(hiddenIndices);

    const angleOfAdjacentBubbles = Math.PI * 2 / NETWORK_DENSITY;
    const distanceOfAdjacentBubbles = Math.hypot(1 - Math.cos(angleOfAdjacentBubbles), 0 - Math.sin(angleOfAdjacentBubbles)) * NETWORK_RING_RADIUS;

    let currentRing = 0;
    let indexOfBubbleInRing = 0;
    let totalBubblesInRing = 0;
    let angleInRing = 0;
    for (let i = 0; i < connections.length; i++) {
        // Make a new position object for this connection.
        positions[i] = {
            id: connections[i].id,
            x: 0,
            y: 0
        };

        // If this is a hidden connection, we can leave its position at (0,0).
        if (hiddenIndicesSetClone.delete(i)) continue;

        // If there are no more bubbles to add to this ring, start on the next ring.
        if (indexOfBubbleInRing == totalBubblesInRing) {
            // Increment the current ring index.
            currentRing++;
            // Shorten some var names.
            const D = distanceOfAdjacentBubbles;
            const R = NETWORK_RING_RADIUS * currentRing;
            // Calculate the next X and Y position given a distance around the ring and a ring
            // radius. This formula was found using Wolfram Alpha and Desmos, so just kinda accept
            // it without understanding it...
            const nextY = Math.sqrt(4 * D ** 2 * R ** 2 - D ** 4) / (2 * R);
            const nextX = Math.sqrt(R ** 2 - nextY ** 2) * (D < Math.sqrt(2 * R ** 2) ? 1 : -1);
            // From the X and Y position for the next point along the circle, calculate the angle it
            // makes in the ring.
            const angleOfNext = Math.atan2(nextY, nextX);
            // Calculate how many bubbles fit in the ring now that we know the angle between
            // bubbles.
            const maxBubblesInRing = Math.floor(Math.round(Math.PI * 2 / angleOfNext * 1e5) / 1e5);
            // Calculate how many bubbles are left to display.
            const leftoverBubbles = positions.length - i - hiddenIndicesSetClone.size;
            // Use the min of the two previous numbers to determine how many bubbles should actually
            // go in the ring.
            totalBubblesInRing = Math.min(leftoverBubbles, maxBubblesInRing);
            // With the number of bubbles in the ring, we can determine what the actual angle
            // between bubbles will be (`angleOfNext` won't round correctly since it could lead to
            // fractional bubbles, `angleInRing` accounts for this).
            angleInRing = Math.PI * 2 / totalBubblesInRing;
            // Set the index of the current bubble in the ring to 0 since we are starting a new
            // ring.
            indexOfBubbleInRing = 0;
        }

        // The angle of the bubble from the center.
        const bubbleAngle = indexOfBubbleInRing * angleInRing - Math.PI / 2;
        // The distance from the center (determined by the ring index and radius).
        const bubbleDistance = currentRing * NETWORK_RING_RADIUS;

        // Add the final position to the list.
        positions[i].x = Math.round(Math.cos(bubbleAngle) * bubbleDistance * 1e5) / 1e5;
        positions[i].y = Math.round(Math.sin(bubbleAngle) * bubbleDistance * 1e5) / 1e5;

        indexOfBubbleInRing++;
    }

    return positions;
};

/**
 * This function takes in a list of connections and positions, and updates the positions so that the
 * network looks spread out more and somewhat organized as a network should be. Nodes with an edge
 * connecting them are more likely to clump together so that edges don't travel as far, and nodes
 * are also repelled from each other so that they're less likely to be too close together.
 *
 * Connections that appear earlier in the list of connections are slightly more likely to appear
 * closer to the center of the network (which allows us to sort the connections first by whatever
 * order we want), but this force is only slightly increased, so there are still nodes all over the
 * place from different indices.
 *
 * This algorithm was copied and modified from this paper:
 * https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=6297585
 * "Force-Directed Layout of Node-Link Diagrams" section
 *
 * @param {Array<{id: import("../scripts/hardcoded_data").UserID}>} connections The list of
 *      connections to assign positions to. All that's needed in each connection is the User ID.
 * @param {Iterable<Number>} hiddenIndices The list of indices of connections that should be hidden. As
 *      in `assignRingPositions`, these connections will end up with a position of (0, 0).
 * @returns {Array<{x: Number, y: Number, id: import("../scripts/hardcoded_data").UserID}>} The list
 *      of positions for each connection. The order of each position will be in the same order as
 *      was provided for `connections`, but each object will still have an ID assigned to it so that
 *      that can be used instead.
 */
const updatePositions = (connections, hiddenIndices, previousPositions) => {
    // Some constants that determine spacing between bubbles.

    // The spring constant between adjacent nodes. Adjacent nodes will try to be be this far apart
    // with their neighbors. Since they may be connected to multiple at once though, it won't be
    // exactly this distance.
    const L = NETWORK_RING_RADIUS;
    // The repulsive force constant. This determines how much nearby nodes will repel each other.
    const K_r = NETWORK_RING_RADIUS ** 3.2;
    // The spring constant that determines how stiff the "spring" between adjacent nodes is. A
    // higher spring constant tries harder to keep nodes at the specified `L` length whereas lower
    // numbers allow for more flexibility.
    const K_s = 6;
    // How fast the bubbles converge on each other. A smaller time step is more accurate and
    // prevents nodes from overshooting their desired position, but takes more iterations to get
    // right.
    const delta_t = .05;
    // The total number of nodes we're looking through.
    const N = connections.length;
    // The maximum displacement one iteration should allow for (squared).
    const MAX_DISPLACEMENT_SQUARED = BUBBLE_SIZE ** 1.5;

    const positions = new Array(connections.length);
    const hiddenIndicesSetClone = new Set(hiddenIndices);

    // Create a new object where we can get the position index given an ID.
    const IDToPositionIndex = {};
    for (let i = 0; i < previousPositions.length; i++) {
        IDToPositionIndex[previousPositions[i].id] = i;
    }

    // Create a new object where we can index each connection by its ID instead of its index in
    // `connections`.
    const IDToConnectionIndex = {};
    for (let i = 0; i < connections.length; i++) {
        IDToConnectionIndex[connections[i].id] = i;
    }

    // Get a set of all the self user's friends.
    const setOfSelfFriends = new Set(selfUser.friends);

    // Since we need to reference connections by their IDs instead of their index within
    // `connections` at certain parts, we need to transform `hiddenIndices` to a set of connections.
    // That way we can just check if a connection is within that set in O(1) without having to know
    // its index.
    const hiddenConnections = new Set(Array.from(hiddenIndices).map(index => connections[index]));

    const forces = new Array(N);

    // Assign positions based on previousPositions.
    for (let i = 0; i < connections.length; i++) {
        // Assign an initial position of (0,0).
        positions[i] = {
            id: connections[i].id,
            x: 0,
            y: 0
        };
        const previousPosition = previousPositions[IDToPositionIndex[connections[i].id]] || { x: 0, y: 0};

        if (hiddenIndicesSetClone.has(i)) {
            // Keep coordinates at (0, 0).
        } else if (previousPosition.x == 0 && previousPosition.y == 0) {
            // If the previous position was at (0,0), but the bubble is not currently being hidden,
            // we perturb the bubbles position so that it will repel away from the center instead of
            // lying directly on top of the self bubble.
            const perturbAngle = seeded_rand(connections[i].id) * Math.PI * 2;
            const perturbDistance = BUBBLE_SIZE / 2;
            positions[i].x = Math.cos(perturbAngle) * perturbDistance;
            positions[i].y = Math.sin(perturbAngle) * perturbDistance;
        } else {
            // Otherwise, we can just copy over the previous position.
            positions[i].x = previousPosition.x;
            positions[i].y = previousPosition.y;
        }
    }

    // Do 250 iterations of making small adjustments to positions.
    for (let iteration = 0; iteration < 250; iteration++) {
        // Initialize the force on each bubble to 0.
        for (let i = 0; i < N; i++) {
            forces[i] = {
                x: 0,
                y: 0
            };
        }

        // Calculate how much repulsion each node needs to do on every other node.
        for (let i1 = 0; i1 < N; i1++) {
            const node1 = connections[i1];
            // Skip over nodes that are hidden away since we don't want their position at (0,0) to
            // affect other nodes (since they're supposed to be hidden away off the graph).
            if (hiddenConnections.has(node1)) continue;

            // This for loop only loops through each edge once even though they are undirected and
            // go both directions.
            for (let i2 = i1 + 1; i2 < N; i2++) {
                const node2 = connections[i2];
                // Skip again if the second node is hidden.
                if (hiddenConnections.has(node2)) continue;

                // Calculate the difference in distance.
                const dx = positions[i2].x - positions[i1].x;
                const dy = positions[i2].y - positions[i1].y;

                // Only make an adjustment if the distance between the two nodes is not 0 since
                // having a distance of 0 would cause issues with dividing by 0.
                if (dx != 0 || dy != 0) {
                    const actualDistance = Math.hypot(dx, dy);
                    const distance = Math.max(actualDistance - BUBBLE_SIZE, 1);
                    const force = K_r / (distance * distance);
                    const fx = force * dx / distance;
                    const fy = force * dy / distance;
                    forces[i1].x -= fx;
                    forces[i1].y -= fy;
                    forces[i2].x += fx;
                    forces[i2].y += fy;
                }
            }

            // Since we looped through the list of connections, all the other nodes had their
            // repulsion calculated relative to each other, but we haven't yet include a repulsion
            // from the self bubble. Do that here.
            const dx = -positions[i1].x;
            const dy = -positions[i1].y;
            if (dx != 0 || dy != 0) {
                const actualDistance = Math.hypot(dx, dy);
                const distance = Math.max(actualDistance - BUBBLE_SIZE, 1);
                const force = K_r / (distance * distance);
                const fx = force * dx / distance;
                const fy = force * dy / distance;
                forces[i1].x -= fx;
                forces[i1].y -= fy;
            }
        }

        // Calculate the spring force between each pair of adjacent nodes.
        for (const i1 in IDToConnectionIndex) {
            const index1 = IDToConnectionIndex[i1];
            const node1 = connections[index1];
            // Skip if this node is hidden.
            if (hiddenConnections.has(node1)) continue;

            // Go through all the friends for this connection to see if they are in the network.
            // This can be improved later on if we need so that we pre-compute mutual friends
            // instead of doing it again on each of the 500 iterations.
            for (let j = 0; j < user_pool[connections[index1].id].friends.length; j++) {
                const i2 = user_pool[connections[index1].id].friends[j];
                // We only care if their friend is also a connection, since otherwise, they're not
                // actually in the network to begin with.
                if (!(i2 in IDToConnectionIndex)) continue;
                const index2 = IDToConnectionIndex[i2];
                const node2 = connections[index2];
                // Skip if this connection is hidden.
                if (hiddenConnections.has(node2)) continue;

                // Only do this once for each pair of adjacent nodes.
                if (index1 < index2) {
                    // Calculate distance between nodes.
                    const dx = positions[index2].x - positions[index1].x;
                    const dy = positions[index2].y - positions[index1].y;
                    // Only calculate if we won't be dividing by 0.
                    if (dx != 0 || dy != 0) {
                        // Calculate a force for this spring. The further the distance is from the
                        // resting spring position, the more force is applied, which encourages
                        // adjacent nodes to be next to each other.
                        const actualDistance = Math.hypot(dx, dy);
                        const distance = Math.max(actualDistance - BUBBLE_SIZE, 1);
                        const force = K_s * (distance - L);
                        const fx = force * dx / distance;
                        const fy = force * dy / distance;
                        forces[IDToConnectionIndex[i1]].x += fx;
                        forces[IDToConnectionIndex[i1]].y += fy;
                        forces[IDToConnectionIndex[i2]].x -= fx;
                        forces[IDToConnectionIndex[i2]].y -= fy;
                    }
                }
            }

            // Do the same as above, but adding a spring between the self bubble and the node now
            // (but only if this connection is in the self user's list of friends).
            if (setOfSelfFriends.has(+i1)) {
                const dx = -positions[index1].x;
                const dy = -positions[index1].y;
                if (dx != 0 || dy != 0) {
                    const actualDistance = Math.hypot(dx, dy);
                    const distance = Math.hypot(actualDistance - BUBBLE_SIZE, 1);
                    const force = K_s * (distance - L);
                    const fx = force * dx / distance;
                    const fy = force * dy / distance;
                    // Multiply the force slightly depending on the index of the connection. This
                    // makes connections that appear earlier in the list pulled in more towards the
                    // center. This not only makes more recent connections more likely to appear in
                    // the center, but also allows the network to grow outwards further so that
                    // nodes don't just clump in the center.
                    forces[IDToConnectionIndex[i1]].x += fx * (2 - IDToConnectionIndex[i1] / N);
                    forces[IDToConnectionIndex[i1]].y += fy * (2 - IDToConnectionIndex[i1] / N);
                }
            }
        }

        // Using the forces we calculated from repulsion and springs, update the positions of each
        // connection.
        for (let i = 0; i < N; i++) {
            const node = connections[i];
            // Skip if this node is hidden since we want it to remain at (0,0).
            if (hiddenConnections.has(node)) continue;

            // Find the distance in each direction we need to move.
            const dx = delta_t * forces[i].x;
            const dy = delta_t * forces[i].y;
            const displacementSquared = dx * dx + dy * dy;
            // If the calculated displacement is too great for just one time step (as determined by
            // `MAX_DISPLACEMENT_SQUARED`), lower its displacement so we don't jump around too much.
            if (displacementSquared > MAX_DISPLACEMENT_SQUARED) {
                const s = Math.sqrt(MAX_DISPLACEMENT_SQUARED / displacementSquared);
                dx = dx * s;
                dy = dy * s;
            }
            // Update the X and Y positions.
            positions[i].x += dx;
            positions[i].y += dy;
        }
    }

    return positions;
}

/**
 * The graphical network that is shown by default. This shows the self user in the center and all
 * their connection expanding outwards.
 * @param {object} props
 * @param {Array<import("../scripts/hardcoded_data").Connection>} props.allConnections
 * @param {Array<Number>} props.hiddenIndices
 * @param {React.MutableRefObject<TextInput>} props.searchBarRef
 */
const GraphicalNetwork = ({allConnections, hiddenIndices, searchBarRef, ...props}) => {
    // Animated values for panning and pinching on the network.
    const dragPosition = useMemo(() => new Animated.ValueXY(), []);
    const offsetZoomScale = useMemo(() => new Animated.Value(1), []);
    const baseZoomScale = useMemo(() => new Animated.Value(1), []);
    const baseZoomScalePlain = useRef(1);
    const zoomScale = useMemo(() => Animated.multiply(offsetZoomScale, baseZoomScale).interpolate({
        inputRange: [MIN_ZOOM, MAX_ZOOM],
        outputRange: [MIN_ZOOM, MAX_ZOOM],
        extrapolate: "clamp"
    }), [offsetZoomScale, baseZoomScale]);

    // Running animations. `momentumAnimation` is the animation when the user swipes on the screen.
    const momentumAnimation = useRef();
    // `returningAnimation` is the animation for when the user taps the self bubble and returns to
    // the center.
    const returningAnimation = useRef();
    // A ref for a boolean indicating whether we are currently centered on the network screen.
    const isCentered = useRef(true);

    // Height and width of the network.
    const [networkWidth, setNetworkWidth] = useState(BUBBLE_SIZE + 1);
    const [networkHeight, setNetworkHeight] = useState(BUBBLE_SIZE + 1);

    const colors = useColors();
    const navigation = useNavigation();

    // `previousPositions` is the positions of each bubble on the previous render. We use this to
    // determine which bubbles change position on the current render so we can animate it moving.
    const previousPositions = useRef(assignInitialPositions(allConnections, hiddenIndices));
    const previousEdgeAngles = useRef({});

    // Assign positions to each bubble.
    // const positions = useMemo(() => assignPositions(allConnections, hiddenIndices, previousPositions), [allConnections, hiddenIndices]);
    const positions = useMemo(() => updatePositions(allConnections, hiddenIndices, previousPositions.current), [allConnections, hiddenIndices]);

    const bubbleAnimatedPositionsById = useRef({});
    const bubbleAnimatedSizesById = useRef({});
    const edgeAnimatedAnglesByIds = useRef({});
    const edgeAnimatedLengthsByIds = useRef({});
    const edgeAnimatedOpacitiesByIds = useRef({});

    // A ref with a boolean indicating whether this is the first render or not.
    const firstRender = useRef(true);
    useEffect(() => firstRender.current = false, []);

    // Generate a map between IDs and the position of the corresponding bubble.
    /** @type {{ [key: import("../scripts/hardcoded_data").UserID]: { x: Number, y: Number } }} */
    const IDToPosition = useMemo(() => {
        const IDToPosition = {};
        for (let i = 0; i < allConnections.length; i++) {
            IDToPosition[allConnections[i].id] = positions[i];
        }
        IDToPosition[selfUser.id] = { x: 0, y: 0 };
        return IDToPosition;
    }, [JSON.stringify(positions)]);

    // Do the same for previous position.
    /** @type {{ [key: import("../scripts/hardcoded_data").UserID]: { x: Number, y: Number } }} */
    const IDToPrevPosition = useMemo(() => {
        const IDToPrevPosition = {};
        for (let i = 0; i < previousPositions.current.length; i++) {
            IDToPrevPosition[previousPositions.current[i].id] = previousPositions.current[i];
        }
        IDToPrevPosition[selfUser.id] = { x: 0, y: 0 };
        return IDToPrevPosition;
    }, [JSON.stringify(previousPositions.current)]);

    // Do the same but for the index within `allConnections`.
    /** @type {{ [key: import("../scripts/hardcoded_data").UserID]: Number }} */
    const IDToConnectionIndex = useMemo(() => {
        const IDToConnectionIndex = {};
        for (let i = 0; i < allConnections.length; i++) {
            IDToConnectionIndex[allConnections[i].id] = i;
        }
        return IDToConnectionIndex;
    }, [JSON.stringify(allConnections)]);

    // Make an Animated.ValueXY for each bubble so we can animate their positions and their sizes.
    /** @type {[Array<Animated.ValueXY>, Array<Animated.AnimatedValue>]} */
    const [bubbleAnimatedPositions, bubbleAnimatedSizes] = useMemo(() => {
        const bubbleAnimatedPositions = new Array(allConnections.length);
        const bubbleAnimatedSizes = new Array(allConnections.length);

        for (let i = 0; i < allConnections.length; i++) {
            const ID = allConnections[i].id;
            const prevPosition = IDToPrevPosition[ID] || { x: 0, y: 0 };

            const position =
                bubbleAnimatedPositionsById.current[ID] ??
                new Animated.ValueXY(
                    firstRender.current ? positions[i] : prevPosition
                );
            bubbleAnimatedPositions[i] = position;
            bubbleAnimatedPositionsById.current[ID] = position;

            const size =
                bubbleAnimatedSizesById.current[ID] ??
                new Animated.Value(
                    prevPosition == 0 && prevPosition == 0 ? 0.2 : 1
                );
            bubbleAnimatedSizes[i] = size;
            bubbleAnimatedSizesById.current[ID] = size;
        }

        return [bubbleAnimatedPositions, bubbleAnimatedSizes];
    }, [JSON.stringify(allConnections)]);

    // Generate the list of edges for the network along with a mapping of ID to the list of edges
    // that involve that ID.
    /** @type {[Array<[import("../scripts/hardcoded_data").UserID, import("../scripts/hardcoded_data").UserID]>, { [key: import("../scripts/hardcoded_data").UserID]: Array<[import("../scripts/hardcoded_data").UserID, import("../scripts/hardcoded_data").UserID]>] }]} */
    const edges = useMemo(() => {
        const edges = [];

        const IDsInNetwork = new Set(allConnections.map(connection => connection.id));

        // First add edges between the self user and their friends.
        for (const friendID of selfUser.friends) {
            const edge = [selfUser.id < friendID ? selfUser.id : friendID, selfUser.id < friendID ? friendID : selfUser.id];
            edges.push(edge);
        }

        // Now add connections between other people in the network.
        for (const connection of allConnections) {
            for (const friendID of user_pool[connection.id].friends) {
                // Continue if this person's friend is not in the network or the friend is the self
                // user.
                if (!IDsInNetwork.has(friendID) || friendID == selfUser.id) continue;
                if (connection.id < friendID) {
                    edges.push([connection.id, friendID]);
                }
            }
        }

        return edges;
    }, [JSON.stringify(allConnections)]);

    // Generate animated value corresponding to each edge controlling the angle the edge makes, its
    // length, and its opacity.
    /** @type {[Array<Animated.AnimatedValue>, Array<Animated.AnimatedValue>, Array<Animated.AnimatedValue>]} */
    const [edgeAnimatedAngles, edgeAnimatedLengths, edgeAnimatedOpacities] = useMemo(() => {
        const edgeAngles = new Array(edges.length);
        const edgeLengths = new Array(edges.length);
        const edgeOpacities = new Array(edges.length);

        for (let i = 0; i < edges.length; i++) {
            const edge = edges[i];
            const posStart = IDToPosition[edge[0]];
            const posEnd = IDToPosition[edge[1]];
            const calculatedEdgeLength = Math.hypot(posEnd.y - posStart.y, posEnd.x - posStart.x);

            const edgeAngle =
                edgeAnimatedAnglesByIds.current[edge[0]]?.[edge[1]] ??
                new Animated.Value(
                    Math.atan2(posEnd.y - posStart.y, posEnd.x - posStart.x)
                );
            edgeAngles[i] = edgeAngle;
            edgeAnimatedAnglesByIds.current[edge[0]] = edgeAnimatedAnglesByIds.current[edge[0]] || {};
            edgeAnimatedAnglesByIds.current[edge[0]][edge[1]] = edgeAngle;

            const edgeLength =
                edgeAnimatedLengthsByIds.current[edge[0]]?.[edge[1]] ??
                new Animated.Value(calculatedEdgeLength);
            edgeLengths[i] = edgeLength;
            edgeAnimatedLengthsByIds.current[edge[0]] = edgeAnimatedLengthsByIds.current[edge[0]] || {};
            edgeAnimatedLengthsByIds.current[edge[0]][edge[1]] = edgeLength;

            const edgeOpacity =
                edgeAnimatedOpacitiesByIds.current[edge[0]]?.[edge[1]] ??
                new Animated.Value(
                    edge[0] != selfUser.id && edge[1] != selfUser.id && ((posStart.x == 0 && posStart.y == 0) || (posEnd.x == 0 && posEnd.y == 0)) ?
                    0 :
                    EDGE_OPACITY(calculatedEdgeLength)
                );
            edgeOpacities[i] = edgeOpacity;
            edgeAnimatedOpacitiesByIds.current[edge[0]] = edgeAnimatedOpacitiesByIds.current[edge[0]] || {};
            edgeAnimatedOpacitiesByIds.current[edge[0]][edge[1]] = edgeOpacity;
        }

        return [edgeAngles, edgeLengths, edgeOpacities];
    }, [JSON.stringify(edges)]);

    // Generate the components that will represent each edge.
    const edgeComponents = useMemo(() => {
        const components = [];

        // Go through each edge and make a line to connect the two nodes.
        for (let i = 0; i < edges.length; i++) {
            components.push(
                <Animated.View
                    key={edges[i].toString()}
                    style={{
                        opacity: edgeAnimatedOpacities[i],
                        width: 1,
                        height: 3,
                        backgroundColor: colors.FLAT1,
                        position: "absolute",
                        top: BUBBLE_SIZE / 2,
                        left: BUBBLE_SIZE / 2,
                        transform: [
                            { translateX: edges[i][0] == selfUser.id ? 0 : bubbleAnimatedPositions[IDToConnectionIndex[edges[i][0]]].x },
                            { translateY: edges[i][0] == selfUser.id ? 0 : bubbleAnimatedPositions[IDToConnectionIndex[edges[i][0]]].y },
                            { translateX: -1 / 2 },
                            { rotateZ: edgeAnimatedAngles[i].interpolate({inputRange: [0,1], outputRange: ["0rad", "1rad"]}) },
                            { translateX: Animated.divide(edgeAnimatedLengths[i], 2) },
                            { scaleX: edgeAnimatedLengths[i] }
                        ]
                    }}
                />
            );
        }

        return components;
    }, [JSON.stringify(edges), JSON.stringify(positions), colors.FLAT1]);

    // This `useMemo` is really a `useEffect` since there's no value returned. The reason it's
    // `useMemo` instead is because `useEffect` is run after the current render is done. We need the
    // code inside to run synchronously before the code that follows it, but we only want it to run
    // when `position` has changed, so we can't just run it inline. If you change it to `useEffect`,
    // it completely breaks animations because `previousPositions` will be updated before the code
    // gets run, and by then, we won't be able to detect changes in positions, so it has to be run
    // synchronously in `useMemo` instead.
    void useMemo(() => {
        /** @type {Set<import("../scripts/hardcoded_data").UserID>} */
        const movingBubbles = new Set();

        // Go through and look for bubbles that changed their positions. For each one, animate their
        // positions moving.

        // We compare against an EPSILON to see if the bubble has moved more than that instead of
        // comparing directly with ==. Positions tend to move around by a millionth of a pixel or
        // so, and we obviously don't need to animated a bubble moving basically to the exact same
        // spot, so we use this epsilon value to have a very tiny cutoff that we consider zero
        // movement.
        const EPSILON = 1e-4;
        for (let i = 0; i < positions.length; i++) {
            if (EPSILON < Math.abs(positions[i].x - (previousPositions.current[i]?.x ?? 0)) || EPSILON < Math.abs(positions[i].y - (previousPositions.current[i]?.y ?? 0))) {
                movingBubbles.add(allConnections[i].id);
                // Start an animation for moving the bubble.
                Animated.timing(
                    bubbleAnimatedPositions[i],
                    {
                        toValue: positions[i],
                        duration: BUBBLE_MOVEMENT_DURATION,
                        easing: Easing.out(Easing.ease),
                        isInteraction: false,
                        useNativeDriver: true
                    }
                ).start();

                if (positions[i].x == 0 && positions[i].y == 0) {
                    // If the bubble is moving towards (0,0), we can assume it's being hidden away.
                    // Animate the size shrinking as it goes towards the center to further indicate
                    // that it's going away.
                    Animated.timing(
                        bubbleAnimatedSizes[i],
                        {
                            toValue: 0.2,
                            duration: BUBBLE_MOVEMENT_DURATION,
                            easing: Easing.out(Easing.ease),
                            isInteraction: false,
                            useNativeDriver: true
                        }
                    ).start();
                } else if ((previousPositions.current[i]?.x ?? 0) == 0 && (previousPositions.current[i]?.y ?? 0) == 0) {
                    // Do the same thing for if the bubble is leaving away from (0,0) (i.e., it's
                    // "appearing", so its size needs to grow back to normal).
                    Animated.timing(
                        bubbleAnimatedSizes[i],
                        {
                            toValue: 1,
                            duration: BUBBLE_MOVEMENT_DURATION,
                            easing: Easing.out(Easing.ease),
                            isInteraction: false,
                            useNativeDriver: true
                        }
                    ).start();
                }
            }
        }

        for (let i = 0; i < edges.length; i++) {
            if (movingBubbles.has(edges[i][0]) || movingBubbles.has(edges[i][1])) {
                const posStart = edges[i][0] == selfUser.id ? { x: 0, y: 0 } : positions[IDToConnectionIndex[edges[i][0]]];
                const posEnd = edges[i][1] == selfUser.id ? { x: 0, y: 0 } : positions[IDToConnectionIndex[edges[i][1]]];
                let jumpToAngle = false;
                if (
                    !(edges[i][0] in IDToPrevPosition) ||
                    !(edges[i][1] in IDToPrevPosition) ||
                    (edges[i][0] != selfUser.id && IDToPrevPosition[edges[i][0]].x == 0 && IDToPrevPosition[edges[i][0]].y == 0 && posStart.x != 0 && posStart.y != 0 && posEnd.x == 0 && posEnd.y == 0) ||
                    (edges[i][1] != selfUser.id && IDToPrevPosition[edges[i][1]].x == 0 && IDToPrevPosition[edges[i][1]].y == 0 && posEnd.x != 0 && posEnd.y != 0 && posStart.x == 0 && posStart.y == 0)
                ) {
                    jumpToAngle = true;
                }
                const edgeAngle = Math.atan2(posEnd.y - posStart.y, posEnd.x - posStart.x);
                const edgeLength = Math.hypot(posEnd.y - posStart.y, posEnd.x - posStart.x);
                const prevEdgeAngle = jumpToAngle ? edgeAngle : previousEdgeAngles.current[edges[i][0]]?.[edges[i][1]] || 0;
                const updatedEdgeAngle = Math.round((prevEdgeAngle - edgeAngle) / (2 * Math.PI)) * 2 * Math.PI + edgeAngle;
                previousEdgeAngles.current[edges[i][0]] = previousEdgeAngles.current[edges[i][0]] || {};
                previousEdgeAngles.current[edges[i][0]][edges[i][1]] = updatedEdgeAngle;

                // If the edge length is heading towards 0, there's no need to animate any change in
                // the angle since the end angle won't be visible anyway.
                if (edgeLength) {
                    if (jumpToAngle) {
                        edgeAnimatedAngles[i].setValue(edgeAngle);
                    } else {
                        Animated.timing(
                            edgeAnimatedAngles[i],
                            {
                                toValue: updatedEdgeAngle,
                                duration: BUBBLE_MOVEMENT_DURATION,
                                easing: Easing.out(Easing.ease),
                                isInteraction: false,
                                useNativeDriver: true
                            }
                        ).start();
                    }
                }
                Animated.timing(
                    edgeAnimatedLengths[i],
                    {
                        toValue: edgeLength,
                        duration: BUBBLE_MOVEMENT_DURATION,
                        easing: Easing.out(Easing.ease),
                        isInteraction: false,
                        useNativeDriver: true
                    }
                ).start();
                Animated.timing(
                    edgeAnimatedOpacities[i],
                    {
                        toValue:
                            edges[i][0] != selfUser.id && edges[i][1] != selfUser.id && ((posStart.x == 0 && posStart.y == 0) || (posEnd.x == 0 && posEnd.y == 0)) ?
                            0 :
                            EDGE_OPACITY(edgeLength),
                        duration: BUBBLE_MOVEMENT_DURATION / 2,
                        easing: Easing.out(Easing.ease),
                        isInteraction: false,
                        useNativeDriver: true
                    }
                ).start();
            }
        }
    }, [JSON.stringify(positions)]);

    previousPositions.current = positions;

    return (
        <PanGestureHandler
            onHandlerStateChange={({nativeEvent: {state, x, y, velocityX, velocityY}}) => {
                switch (state) {
                    case State.BEGAN:
                        // Stop any running animations.
                        momentumAnimation.current?.stop();
                        momentumAnimation.current = null;
                        returningAnimation.current?.stop();
                        returningAnimation.current = null;
                        dragPosition.extractOffset();
                        break;
                    case State.END:
                        dragPosition.flattenOffset();
                        momentumAnimation.current = Animated.decay(
                            dragPosition,
                            {
                                // React Native Gesture Handler returns velocity in points per
                                // second while React Native Animated interprets velocity as points
                                // per MILLIsecond for some reason (???). Divide both velocities by
                                // 1000 here to convert between the two.
                                // https://github.com/facebook/react-native/issues/18953
                                velocity: {
                                    x: velocityX / 1000,
                                    y: velocityY / 1000
                                },
                                deceleration: .995,
                                useNativeDriver: true
                            }
                        );
                        momentumAnimation.current.start();
                        isCentered.current = false;
                        break;
                }
            }}
            onGestureEvent={Animated.event(
                [{ nativeEvent: { translationX: dragPosition.x, translationY: dragPosition.y } }],
                { useNativeDriver: true }
            )}
        >
            <Animated.View
                style={{
                    flexGrow: 1
                }}
                onStartShouldSetResponderCapture={() => {
                    if (searchBarRef.current?.isFocused()) {
                        searchBarRef.current?.blur();
                        return true;
                    }
                    return false;
                }}
            >
                <PinchGestureHandler
                    onHandlerStateChange={({nativeEvent: {oldState, scale}}) => {
                        // Once the user is done zooming, save the current zoom scale to another
                        // value so we can start a new zoom starting from the current zoom scale.
                        if (oldState == State.ACTIVE) {
                            baseZoomScalePlain.current = Math.min(Math.max(scale * baseZoomScalePlain.current, MIN_ZOOM), MAX_ZOOM);
                            baseZoomScale.setValue(baseZoomScalePlain.current);
                            offsetZoomScale.setValue(1);
                        }
                    }}
                    onGestureEvent={Animated.event(
                        [{ nativeEvent: { scale: offsetZoomScale } }],
                        { useNativeDriver: true }
                    )}
                >
                    {/* The <View> that encompasses the entire network. It's an Animated.View so that the
                        parent PanGestureHandler's onGestureEvent can be bound to Animated.event instead
                        of a function callback. */}
                    <Animated.View
                        style={{
                            flexGrow: 1,
                            justifyContent: "center",
                            alignItems: "center",
                            overflow: "hidden"
                        }}
                        onLayout={({nativeEvent: {layout: {height, width}}}) => {
                            setNetworkHeight(height);
                            setNetworkWidth(width)
                        }}
                    >
                        {/* This <View> is the parent of all the network bubbles inside the network. It has
                            a transformation applied to it that allows the user to scroll through the
                            network by translating this element instead of all the individual bubbles. */}
                        <Animated.View
                            style={{
                                height: BUBBLE_SIZE,
                                width: BUBBLE_SIZE,
                                transform: [
                                    { translateX: dragPosition.x },
                                    { translateY: dragPosition.y },
                                    { scale: zoomScale }
                                ]
                            }}
                        >
                            {edgeComponents}
                            {
                                allConnections.map((connection, i) =>
                                    <NetworkBubble
                                        key={connection.id}
                                        user={user_pool[connection.id]}
                                        networkHeight={networkHeight}
                                        networkWidth={networkWidth}
                                        dragPosition={dragPosition}
                                        zoomScale={zoomScale}
                                        animatedPosition={bubbleAnimatedPositions[i]}
                                        animatedSize={bubbleAnimatedSizes[i]}
                                        onPress={e => {
                                            navigation.navigate("Other's Profile Screen", {
                                                userID: connection.id
                                            });
                                        }}
                                    />
                                )
                            }
                            <NetworkBubble
                                user={selfUser}
                                networkHeight={networkHeight}
                                networkWidth={networkWidth}
                                dragPosition={dragPosition}
                                zoomScale={zoomScale}
                                animatedPosition={{ x: 0, y: 0 }}
                                animatedSize={1}
                                onPress={e => {
                                    if (isCentered.current) {
                                        // If we are already at the center of the network, return to
                                        // zoom scale 1.
                                        // Use a spring animation for maximum fun-ness ;).
                                        Animated.spring(
                                            baseZoomScale,
                                            {
                                                toValue: 1,
                                                useNativeDriver: true
                                            }
                                        ).start();
                                        baseZoomScalePlain.current = 1;
                                    } else {
                                        // Otherwise, return the network to (0,0).
                                        dragPosition.flattenOffset();
                                        returningAnimation.current = Animated.spring(
                                            dragPosition,
                                            {
                                                toValue: { x: 0, y: 0},
                                                useNativeDriver: true
                                            }
                                        );
                                        returningAnimation.current.start(finished => isCentered.current = finished || isCentered.current);
                                    }
                                }}
                            />
                        </Animated.View>
                        {/* The Add button in the bottom right corner of the network screen. */}
                        <Pressable
                            style={{
                                position: "absolute",
                                bottom: style_values.screen_padding.vertical,
                                right: style_values.screen_padding.horizontal,
                                height: 60,
                                width: 60,
                                borderRadius: 30,
                                overflow: "hidden",
                                justifyContent: "center",
                                alignItems: "center"
                            }}
                            onPress={e => navigation.navigate("Add to Network")}
                        >
                            <LinearGradient
                                style={{
                                    position: "absolute",
                                    top: 0,
                                    left: 0,
                                    height: "100%",
                                    width: "100%"
                                }}
                                colors={[colors.FLAT1, colors.FLAT2]}
                                start={[1.3, .5]}
                                end={[-.1, .5]}
                            />
                            <Icon color="white" height={32}>add</Icon>
                        </Pressable>
                        {/* The gradient at the top of the network screen that prevents a hard cutoff
                            between the network and the search bar. */}
                        <LinearGradient
                            colors={[
                                colors.BACKGROUND,
                                colors.BACKGROUND + "00"
                            ]}
                            style={{
                                width: "100%",
                                height: BUBBLE_SIZE * .2,
                                position: "absolute",
                                top: 0,
                                left: 0
                            }}
                            pointerEvents="none"
                        />
                        {/* The same as the above <LinearGradient>, but for the bottom of the screen.
                        <LinearGradient
                            colors={[
                                colors.BACKGROUND + "00",
                                colors.BACKGROUND
                            ]}
                            style={{
                                width: "100%",
                                height: BUBBLE_SIZE * .2,
                                position: "absolute",
                                bottom: 0,
                                left: 0
                            }}
                            pointerEvents="none"
                        /> */}
                    </Animated.View>
                </PinchGestureHandler>
            </Animated.View>
        </PanGestureHandler>
    );
};

/**
 * A single bubble in the network.
 * @param {object} props
 * @param {import("../scripts/hardcoded_data").User} props.user The user this bubble is for.
 * @param {Animated.AnimatedValueXY} props.dragPosition The animated value controlling the Z and Y
 *      positions for panning.
 * @param {Animated.AnimatedInterpolation} props.zoomScale The animated value controlling the zoom
 *      scale for the network.
 * @param {Animated.AnimatedValueXY} props.animatedPosition The animated value controlling the
 *      position of the bubble.
 * @param {Animated.AnimatedValue} props.animatedSize The animated value controlling the relative
 *      size of the bubble.
 * @param {Number} props.networkHeight The height of the network area. Used to determine where the
 *      edges of the network are.
 * @param {Number} props.networkWidth The width of the network area. Used to determine where the
 *      edges of the network are.
 * @param {object} props.style Any additional style to apply to the bubble's outer view.
 */
const NetworkBubble = ({user, dragPosition, zoomScale, animatedPosition, animatedSize, networkHeight, networkWidth, style, ...props}) => {
    const colors = useColors();
    const AnimatedPressable = useMemo(() => Animated.createAnimatedComponent(TouchableOpacity), []);

    const styles = [
        {
            height: BUBBLE_SIZE,
            width: BUBBLE_SIZE,
            borderRadius: BUBBLE_SIZE / 2,
            backgroundColor: colors.FLAT1,
            padding: 3
        },
        style
    ];

    const image = <Image
        source={{ uri: user.profileImage}}
        style={{
            height: "100%",
            width: "100%",
            borderRadius: BUBBLE_SIZE
        }}
    />

    // Check if this is our own bubble or another person's bubble.
    if (user.id == selfUser.id) {
        // The self bubble moves around with the rest of the network, but when it reached the edges
        // of the screen, it doesn't go completely off screen and instead sticks to the edge. This
        // only happens for the self bubble. To stick to the edge, it has a transformation applied
        // to it that undoes the transformation on the rest of the network, so that even when the
        // rest of the network is moving, the self bubble will stay in place. `xMinPos` and
        // `yMinPos` are for correcting the transformation when the bubble is on the left or top of
        // the screen (i.e., it controls the minimum values for the x and y coordinates) while the
        // `xMaxPos` and `yMaxPos` is for the bottom and right edges of the screen. It's broken up
        // into two because while the bubble is in the center of the screen, we don't want to apply
        // any transformation at all, so each transformation can only activate when the bubble is on
        // one side of the screen, requiring four transformation in total.
        const xMinPos = Animated.divide(dragPosition.x.interpolate({
            inputRange: [ -networkWidth / 2, -networkWidth / 2 + BUBBLE_SIZE * .35 ],
            outputRange: [ BUBBLE_SIZE * .35, 0 ],
            extrapolateRight: "clamp"
        }), zoomScale);
        const yMinPos = Animated.divide(dragPosition.y.interpolate({
            inputRange: [ -networkHeight / 2, -networkHeight / 2 + BUBBLE_SIZE * .35 ],
            outputRange: [ BUBBLE_SIZE * .35, 0 ],
            extrapolateRight: "clamp"
        }), zoomScale);
        const xMaxPos = Animated.divide(dragPosition.x.interpolate({
            inputRange: [ networkWidth / 2 - BUBBLE_SIZE * .35, networkWidth / 2 ],
            outputRange: [ 0, -BUBBLE_SIZE * .35 ],
            extrapolateLeft: "clamp"
        }), zoomScale);
        const yMaxPos = Animated.divide(dragPosition.y.interpolate({
            inputRange: [ networkHeight / 2 - BUBBLE_SIZE * .35, networkHeight / 2 ],
            outputRange: [ 0, -BUBBLE_SIZE * .35 ],
            extrapolateLeft: "clamp"
        }), zoomScale);

        // The self bubble is the only one that's purple.
        styles[0].backgroundColor = colors.FLAT2;

        return (
            <Animated.View style={{
                transform: [
                    { translateX: xMinPos },
                    { translateY: yMinPos }
                ],
                position: "absolute"
            }}>
                <AnimatedPressable
                    style={[
                        {
                            transform: [
                                { translateX: xMaxPos },
                                { translateY: yMaxPos }
                            ]
                        },
                        styles
                    ]}
                    {...props}
                >
                    {image}
                </AnimatedPressable>
            </Animated.View>
        );
    } else {
        return (
            <AnimatedPressable
                style={[
                    {
                        position: "absolute",
                        transform: [
                            { translateX: animatedPosition.x },
                            { translateY: animatedPosition.y },
                            { scale: animatedSize }
                        ],
                    },
                    styles
                ]}
                {...props}
            >
                {image}
            </AnimatedPressable>
        );
    }
}

/**
 * The portion of the screen below the network dedicated to the draggable timeline. This includes
 * the text above the timeline, the draggable colored bar, and the tick marks below the bar.
 * @param {object} props
 */
const DraggableTimeline = ({timeDelta, onChangeTimeDelta, ...props}) => {
    const colors = useColors();

    const now = new Date();
    // Calculate how many years and months in the past the current time is set to. For all the other
    // units (e.g. month, days, hours) in the past, we can just use math to figure it out, but years
    // are different because each year can have 365 or 366 days. 365.25 is a good generalization to
    // make up for that, but if you go back far enough, it starts to mess up still, so we just count
    // the number of years here instead.
    const year = now.getFullYear();
    const setDate = new Date(now.valueOf() - timeDelta);
    let yearsAgo = now.getFullYear() - setDate.getFullYear();
    const setDateThisYear = new Date(setDate.valueOf());
    setDateThisYear.setFullYear(year);
    if (setDateThisYear > now) {
        setDateThisYear.setFullYear(year - 1);
        yearsAgo--;
    }
    let monthsModYearMilliseconds = now - setDateThisYear;
    let monthsModYear = Math.floor(monthsModYearMilliseconds / 1000 / 60 / 60 / 24 / (365/12));
    if (monthsModYear == 12) {
        yearsAgo++;
        monthsModYear = 0;
    }

    // The time that gets shown above the timeline to indicate more precisely how far back it's
    // scrolled to right now.
    const displayedTime =
        timeDelta < 1000 * 60 * 5 ? "Current" :
        timeDelta < 1000 * 60 * 60 ? unitify(Math.floor(timeDelta / 1000 / 60), "minute ago", "minutes ago") :
        timeDelta < 1000 * 60 * 60 * 24 ? unitify(Math.floor(timeDelta / 1000 / 60 / 60), "hour ago", "hours ago") :
        timeDelta < 1000 * 60 * 60 * 24 * 7 ?
            unitify(Math.floor(timeDelta / 1000 / 60 / 60 / 24), "day", "days") + (
                Math.floor((timeDelta / 1000 / 60 / 60) % 24) ?
                ", " + unitify(Math.floor((timeDelta / 1000 / 60 / 60) % 24), "hour ago", "hours ago") :
                " ago"
            ) :
        timeDelta < 1000 * 60 * 60 * 24 * 28 ? unitify(Math.floor(timeDelta / 1000 / 60 / 60 / 24), "day ago", "days ago") :
        timeDelta < 1000 * 60 * 60 * 24 * 30 ? "1 month ago" : // Covers 28-30 days
        timeDelta < 1000 * 60 * 60 * 24 * 365 ?
            unitify(Math.floor(timeDelta / 1000 / 60 / 60 / 24 / 30), "month", "months") + (
                Math.floor(((timeDelta / 1000 / 60 / 60 / 24) % 30) / 7.5) ?
                ", " + unitify(Math.floor(((timeDelta / 1000 / 60 / 60 / 24) % 30) / 7.5), "week", "weeks") :
                ""
            ) + " ago" :
        unitify(yearsAgo, "year", "years") + (
            monthsModYear ?
            ", " + unitify(monthsModYear, "month", "months") :
            ""
        ) + " ago";

    return (
        <View
            style={{
                display: "flex",
                flexDirection: "column",
                alignItems: "stretch",
            }}
        >
            <StyledText style={{textAlign: "center"}}>{displayedTime}</StyledText>
            {/* This <View> is the triangle underneath the text that overlaps on top of the colored
                bar to indicate where along the bar we are. The triangle is just made up of
                borders. */}
            <View
                style={{
                    borderTopWidth: 11,
                    borderLeftWidth: 6,
                    borderRightWidth: 6,
                    borderTopColor: colors.TEXT,
                    borderLeftColor: "transparent",
                    borderRightColor: "transparent",
                    height: 0,
                    width: 0,
                    alignSelf: "center",
                    marginBottom: -6,
                    zIndex: 1
                }}
            />
            <TimelineVisual
                onChangeTimeDelta={onChangeTimeDelta}
            />
        </View>
    );
};

/**
 * The actual timeline that gets shown in a DraggableTimeline parent component. This is the bar with
 * the gradient background and tick marks. It's generated as a FlatList where each "data" unit in
 * the FlatList corresponds to one segment in the timeline (e.g., the segment from 1 day ago to 1
 * week ago). It only goes as far back as the profile's account creation data
 * (`selfUser.creationDate`), but generates a few additional segments farther back in the past that
 * cannot be scrolled to, to giver the illusion that the timeline goes back forever off screen.
 * @param {object} props
 * @param {(Number) => void} onChangeTimeDelta A callback function that gets called each time the
 *      timeline is scrolled. It receives a number as an argument indicating the number of
 *      milliseconds in the past the timeline is scrolled to.
 */
const TimelineVisual = ({onChangeTimeDelta, ...props}) => {
    const SEGMENT_WIDTH = 175;
    const SEGMENT_HEIGHT = 28;

    // Keeps track of the total width of the timeline so we can add spacing before and after the
    // timeline to keep it centered in the screen. This spacing also determines how far back the
    // user can scroll (which should correspond to when their account was created).
    const [totalWidth, setTotalWidth] = useState(Dimensions.get("window").width);

    // Determine how many segments need to be generated to go all the way back to when the user
    // created their account. A user who just finished creating their account ten seconds ago for
    // example will only have to scroll back like a fraction of a pixel, so we can safely generate
    // only 1 segment. A user from two years ago though can scroll back much further since the first
    // few segments cover smaller magnitudes of time. Once we get to the segments that each cover a
    // year, we can calculate it linearly since the segments stop changing magnitudes of time. This
    // is a fractional number. If the user's account is 5 days old for example, we need to generate
    // two FULL segments (now - 1 day ago, and then 1 day ago to 1 week ago), but this calculates
    // how far into the second segment the user will be allowed to scroll.
    const accountAge = new Date() - selfUser.creationDate || 1;
    const fractionalNumberOfSegments =
        accountAge <= 1000 * 60 * 60 * 24 ? accountAge / 1000 / 60 / 60 / 24 :
        accountAge <= 1000 * 60 * 60 * 24 * 7 ? (accountAge / 1000 / 60 / 60 / 24 - 1) / 6 + 1 :
        accountAge <= 1000 * 60 * 60 * 24 * 28 ? (accountAge / 1000 / 60 / 60 / 24 - 7) / 21 + 2 :
        accountAge <= 1000 * 60 * 60 * 24 * 30 * 6 ? (accountAge / 1000 / 60 / 60 / 24 / 30 - 1) / 5 + 3 :
        accountAge <= 1000 * 60 * 60 * 24 * 365 ? (accountAge / 1000 / 60 / 60 / 24 / 30 - 6) / 6 + 4 :
        (accountAge / 1000 / 60 / 60 / 24 / 365 - 1) + 5;
    // Additional segments are generated to make the timeline scroll farther back off screen. We
    // start off with 5 before onLayout has fired, and then use the total width to calculate from
    // there once we know how wide the screen is.
    const additionalSegments = Math.max(5, Math.ceil(totalWidth / SEGMENT_WIDTH));
    // The total number of segments to ber shown.
    const numberOfSegments = Math.ceil(fractionalNumberOfSegments) + additionalSegments;
    const lastIndex = numberOfSegments - 1;
    // An array to use as the data for the FlatList. All we care about is that it has the right
    // length. We can just leave it empty.
    const timelineMagnitudes = useMemo(() => new Array(numberOfSegments), [numberOfSegments]);
    const colors = useColors();

    return (
        <FlatList
            onLayout={({nativeEvent: {layout: {width}}}) => setTotalWidth(width)}
            inverted // Allows us to start at the end and scroll backwards.
            horizontal
            data={timelineMagnitudes}
            keyExtractor={(item, index) => index}
            onScroll={({nativeEvent: {contentOffset: {x}}}) => {
                // Fire `onChangeTimeDelta` each time the user scrolls.

                // Segments is capped at each of the timeline. If the user scrolls even further
                // (which is allowed so long as `bounce` is enabled on the FlatList), it is still
                // capped at where the timeline will scroll back to once they let go.
                const segments = Math.max(0, Math.min(fractionalNumberOfSegments, x / SEGMENT_WIDTH));
                // Determine the number of milliseconds in the past based on the number of segments
                // they are scrolled back to.
                if (segments < 1) {
                    onChangeTimeDelta(segments * 1000 * 60 * 60 * 24);
                } else if (segments < 2) {
                    onChangeTimeDelta((segments - 1) * 1000 * 60 * 60 * 24 * 6 + 1000 * 60 * 60 * 24);
                } else if (segments < 3) {
                    onChangeTimeDelta((segments - 2) * 1000 * 60 * 60 * 24 * 7 * 3 + 1000 * 60 * 60 * 24 * 7);
                } else if (segments < 4) {
                    onChangeTimeDelta((segments - 3) * 1000 * 60 * 60 * 24 * 30 * 5 + 1000 * 60 * 60 * 24 * 30);
                } else if (segments < 5) {
                    onChangeTimeDelta((segments - 4) * 1000 * 60 * 60 * 24 * 30 * 6 + 1000 * 60 * 60 * 24 * 30 * 6);
                } else {
                    onChangeTimeDelta((segments - 4) * 1000 * 60 * 60 * 24 * 365);
                }
            }}
            renderItem={({item, index}) => {
                return (
                    // Each <View> here includes the colored bar (with all its inner tick marks),
                    // as well as a tick mark that extrudes beneath it with some text to identify
                    // how far back that tick mark is.
                    <View
                        style={{
                            display: "flex",
                            flexDirection: "column",
                            // We start off with position: relative, until we reach the additional
                            // segments we want to show in the past, but don't want to affect the
                            // layout. At that point, we switch to position: absolute and position
                            // it ourselves.
                            position:
                                index >= numberOfSegments - additionalSegments ?
                                "absolute" :
                                "relative",
                            left:
                                index >= numberOfSegments - additionalSegments ?
                                (index - (numberOfSegments - additionalSegments)) * SEGMENT_WIDTH + (index == lastIndex ? style_values.border_radius - 2 + 1.5/2 : 0) :
                                0
                        }}
                    >
                        {/* This <View> is the colored bar. It has a linear gradient inside it as
                            well as somet ick marks. */}
                        <View style={{
                            width: SEGMENT_WIDTH + (index == 0 ? style_values.border_radius - 2 - 1.5/2 : 0) + (index == lastIndex ? style_values.border_radius - 2 + 1.5/2 : 0),
                            height: SEGMENT_HEIGHT,
                            borderRadius: style_values.border_radius,
                            borderTopRightRadius: index == 0 ? style_values.border_radius : 0,
                            borderBottomRightRadius: index == 0 ? style_values.border_radius : 0,
                            borderTopLeftRadius: index == lastIndex ? style_values.border_radius : 0,
                            borderBottomLeftRadius: index == lastIndex ? style_values.border_radius : 0,
                            overflow: "hidden"
                        }}>
                            <LinearGradient
                                style={{
                                    position: "absolute",
                                    height: "100%",
                                    width: "100%",
                                    top: 0,
                                    left: 0
                                }}
                                colors={[index == 0 || index == 1 ? colors.FLAT2 : colors.FLAT4, colors.FLAT4]}
                                start={[index == 1 ? 2 : 1, 0.5]}
                                end={[index == 0 ? -1 : 0, 0.5]}
                            />
                            {/* This is the tick mark at the very end of the segment which has a
                                slightly greater width. It also corresponds to the tick mark that
                                extends below the bar. */}
                            <View
                                style={{
                                    height: "100%",
                                    width: 1.5,
                                    backgroundColor: "white",
                                    position: "absolute",
                                    left: SEGMENT_WIDTH - 1.5 + (index == lastIndex ? style_values.border_radius - 2 + 1.5/2 : 0)
                                }}
                            />
                            {/* All the other intermediate tick marks are placed inside this view
                                and spread out evenly using justify-content: space-between. */}
                            <View
                                style={{
                                    display: "flex",
                                    flexDirection: "row",
                                    justifyContent: "space-between",
                                    height: 0.6 * SEGMENT_HEIGHT,
                                    width: SEGMENT_WIDTH - 1.5,
                                    position: "absolute",
                                    top: 0,
                                    left: index == lastIndex ? style_values.border_radius - 2 + 1.5/2 : 0
                                }}
                            >
                                <View style={{width: 0, height: 0}}/>
                                {
                                    // This determines the number of tick marks in each segment and
                                    // how tall each one is as a fraction of the total height.
                                    (
                                        index == 0 ? [.35, .35, .5, .35, .35, .75, .35, .35, .5, .35, .35, 1, .35, .35, .5, .35, .35, .75, .35, .35, .5, .35, .35] :
                                        index == 1 ? [.35, .35, .35, .75, .35, .35, .35, .75, .35, .35, .35, 1, .35, .35, .35, .75, .35, .35, .35, .75, .35, .35, .35] :
                                        index == 2 ? [.35, .35, .35, .35, .35, .35, .75, .35, .35, .35, .35, .35, .35, .75, .35, .35, .35, .35, .35, .35] :
                                        index == 3 ? [.35, .5, .35, 1, .35, .5, .35, 1, .35, .5, .35, 1, .35, .5, .35, 1, .35, .5, .35] :
                                        index == 4 ? [.35, .35, .35, .75, .35, .35, .35, .75, .35, .35, .35, 1, .35, .35, .35, .75, .35, .35, .35, .75, .35, .35, .35] :
                                        [.35, .5, .35, .5, .35, .75, .35, .5, .35, .5, .35, 1, .35, .5, .35, .5, .35, .75, .35, .5, .35, .5, .35]
                                    ).map((tickHeight, i, arr) =>
                                        <View
                                            key={index + i / arr.length}
                                            style={{
                                                width: 1,
                                                height: tickHeight * 100 + "%",
                                                backgroundColor: "white"
                                            }}
                                        />
                                    )
                                }
                                <View style={{width: 0, height: 0}}/>
                            </View>
                            {
                                // We need to place one additional thick tick mark at the end
                                // because there is one extra tick mark compared to the number of
                                // segments.
                                index == lastIndex ?
                                <View
                                    style={{
                                        height: "100%",
                                        width: 1.5,
                                        backgroundColor: "white",
                                        position: "absolute",
                                        left: style_values.border_radius - 2 - 1.5/2
                                    }}
                                />
                                : null
                            }
                        </View>
                        {/* This is the tick mark that extends below the colored bar. It is placed
                            so that it is aligned with the thick tick mark we drew above inside
                            the colored bar. */}
                        <View
                            style={{
                                zIndex: -1,
                                height: 21,
                                width: 1.5,
                                backgroundColor: colors.TEXT,
                                marginTop: -3,
                                position: "relative",
                                right: 1.5,
                                // left: SEGMENT_WIDTH - 1.5 + (index == lastIndex ? style_values.border_radius - 2 + 1.5/2 : 0),
                            }}
                        />
                        {/* This is the associated label for the tick mark above. */}
                        <StyledText
                            style={{
                                position: "absolute",
                                left: SEGMENT_WIDTH + 4 + (index == lastIndex ? style_values.border_radius - 2 : 0),
                                top: 28
                            }}
                        >{
                            index == 0 ? "Now" :
                            index == 1 ? "1 day ago" :
                            index == 2 ? "1 week ago" :
                            index == 3 ? "1 month ago" :
                            index == 4 ? "6 months ago" :
                            index == 5 ? "1 year ago" :
                            (index - 4) + " years ago"
                        }</StyledText>
                        {
                            // Here, we place an additional tick mark and label if this is the last
                            // segment because we need to draw one more tick mark than there are
                            // segments.
                            index == lastIndex ?
                            <>
                                <View
                                    style={{
                                        zIndex: -1,
                                        height: 21,
                                        width: 1.5,
                                        backgroundColor: colors.TEXT,
                                        position: "absolute",
                                        bottom: 0,
                                        left: style_values.border_radius - 2 - 1.5/2
                                    }}
                                />
                                <StyledText
                                    style={{
                                        position: "absolute",
                                        left: style_values.border_radius - 2 - 1.5/2 + 4,
                                        top: 28
                                    }}
                                >{
                                    index == 0 ? "1 day ago" :
                                    index == 1 ? "1 week ago" :
                                    index == 2 ? "1 month ago" :
                                    index == 3 ? "6 months ago" :
                                    index == 4 ? "1 year ago" :
                                    (index - 3) + " years ago"
                                }</StyledText>
                            </>
                            : null
                        }
                    </View>
                )
            }}
            // The ListHeaderComponent is just spacing added to the right of the draggable bar that
            // ensures the bar starts at the center of the screen instead of all the way to the
            // right.
            ListHeaderComponent={<View style={{width: totalWidth / 2 - style_values.border_radius + 2}}/>}
            // The ListFooterComponent is the spacing placed after the (to the left) of the
            // draggable bar. Its width determines how far back along a segment we can scroll to.
            // If the account creation date lines up midway through a segment, you can only scroll
            // partway through that segment since anymore would be in the past before the user's
            // account was created. The width here determines how far back that is.
            ListFooterComponent={<View style={{width: totalWidth / 2 - SEGMENT_WIDTH * (numberOfSegments - additionalSegments - fractionalNumberOfSegments)}}/>}
        />
    )
}

export default NetworkScreen;