import React, { useCallback, useEffect, useMemo, useRef, useState } from "react";
import { HeaderBar } from "../components/HeaderBar";
import { View, ScrollView, TouchableOpacity, Image, Alert, Dimensions } from "react-native";
import StyledText from "../components/StyledText";
import { style_values } from "../scripts/styles";
import usePaddedSafeAreaInsets from "../scripts/usePaddedSafeAreaInsets";
import useColors from "../scripts/useColors";
import Color from "color";
import {selfUser} from '../scripts/hardcoded_data';
import InterestList from "../components/InterestList";
import { ClickableText } from "../components/ClickableText";
import { Animated as AnimatedMapView, AnimatedRegion } from "react-native-maps";
import * as Location from "expo-location";
import useTriggerRerender from "../scripts/useTriggerRerender";
import GlobalEvents from "../scripts/GlobalEvents";

const showUnimplemented = () => {
    Alert.alert(
        ":(",
        "Sorry, this part of the app hasn't been implemented for the prototype.",
        [
            {
                text: "Okay",
                style: "cancel"
            }
        ],
        {
            cancelable: true
        }
    );
};

const SelfProfileScreen = () => {
    const colors = useColors();

    return (
        <View style={{
            display: "flex",
            flexDirection: "column",
            height: "100%",
            backgroundColor: colors.BACKGROUND
        }}>
            <HeaderBar/>
            <SelfProfileBody/>
        </View>
    );
};


const SelfProfileBody = () => {
    const insets = usePaddedSafeAreaInsets();
    const colors = useColors();

    const [isLocationActive, setIsLocationActive] = useState(selfUser.isActive);
    GlobalEvents.useListener("statusupdate", function() {
        setIsLocationActive(selfUser.isActive);
    }, []);

    return (
        <ScrollView contentContainerStyle={{
            paddingLeft: insets.left,
            paddingRight: insets.right,
            paddingTop: style_values.screen_padding.vertical,
            paddingBottom: 10,
        }}>
            <View style={{flexDirection: "row", justifyContent: "space-between", alignItems: "baseline"}}>
                <StyledText type="h2">Your Profile</StyledText>
                <TouchableOpacity
                    onPress={showUnimplemented}
                >
                    <ClickableText onPress={showUnimplemented}>Edit Profile</ClickableText>
                </TouchableOpacity>
            </View>
            <View style={{marginVertical: style_values.screen_padding.vertical, alignItems: "center"}}>
                <Image
                    style={{width: "40%", aspectRatio: 1/1, borderRadius: Dimensions.get("window").width}}
                    source={{uri: selfUser.profileImage}}
                />
                <StyledText style={{fontSize: 20, marginTop: 10}}>{selfUser.firstName} {selfUser.lastName}</StyledText>
            </View>
            <StyledText style={{color: colors.SECONDARY_TEXT, marginBottom: 20, textAlign: 'center'}}>Your profile is visible to <StyledText style={{color: colors.SECONDARY_TEXT, fontWeight: 700}}>everyone</StyledText>.</StyledText>
            <View style={{ marginTop: 20,}}>
                <StyledText type="h3">Your Interests:</StyledText>
                <InterestList
                    interests={selfUser.interests}
                    similar={true}
                />
            </View>
            <View style={{marginTop: 20}}>
                <View>
                    <StyledText type="h3" style={{marginBottom: 5}}>Location:</StyledText>
                    <MapSection isActive={isLocationActive}/>
                </View>
                {
                    isLocationActive ? 
                    <StyledText style={{color: colors.SECONDARY_TEXT, marginTop: 10, textAlign: "center", fontSize: 14}}>Your proximity is visible to <StyledText style={{color: colors.SECONDARY_TEXT, fontWeight: 700, fontSize: 14}}>everyone</StyledText> nearby. Your exact location is not being shared.</StyledText> :
                    <StyledText style={{color: colors.SECONDARY_TEXT, marginTop: 10, textAlign: "center", fontSize: 14}}>Your proximity is not visible to anyone. You will not be able to see others nearby until you become Active again.</StyledText>
                }
                <ClickableText
                    onPress={showUnimplemented}
                    style={{
                        marginTop: 30,
                        color: colors.SECONDARY_TEXT,
                        fontSize: 12,
                        fontWeight: "400",
                        textAlign: "center"
                    }}
                >Learn more about how we use and protect your location data.</ClickableText>
            </View>
        </ScrollView>
    )
};

/**
 * @param {object} props
 * @param {import("react-native").StyleProp<import("react-native").ViewStyle>} props.style
 * @param {boolean} props.isActive
 */
const MapSection = ({style, isActive, ...props}) => {
    const colors = useColors();
    const triggerRerender = useTriggerRerender();

    // Calculate the aspect ratio of the map so that it only takes up a maximum of 40% of the screen
    // height. This is because the user still needs area on the screen to be able to scroll (since
    // scrolling along the map would just move the map instead). The tallest it will get is having
    // an aspect ratio of 1/1.
    const windowAspectRatio = Dimensions.get("window").width / Dimensions.get("window").height;
    const mapAspectRatio = Math.max(1, windowAspectRatio * 2.5);
    const region = useRef(new AnimatedRegion()).current;
    const isLoading = useRef(true);

    const isNotVisible = useRef(true);

    const queryLocation = useCallback(() => {
        Location.getCurrentPositionAsync({ accuracy: Location.Accuracy.Balanced }).then(
            result => {
                if (isNotVisible.current) return;

                if (isLoading.current) {
                    isLoading.current = false;
                    triggerRerender();
                }

                region.timing({
                    latitude: result.coords.latitude,
                    longitude: result.coords.longitude,
                    latitudeDelta: 0.005,
                    longitudeDelta: 0.005,
                    duration: 300,
                    useNativeDriver: true
                }).start();

                setTimeout(queryLocation, 10 * 1000);
            }
        );
    }, []);

    useEffect(() => {
        isNotVisible.current = !isActive;
        if (!isActive) return;
        queryLocation();
        return () => void (isNotVisible.current = true);
    }, [isActive]);

    // Memoize the Map component so it doesn't have to re-render every time.
    const memoizedMap = useMemo(() => {
        if (!isActive || isLoading.current) return null;
        return (
            <AnimatedMapView
                showsUserLocation={true}
                loadingEnabled={true}
                userInterfaceStyle={colors.COLOR_SCHEME}
                region={region}
                scrollEnabled={false}
                rotateEnabled={false}
                zoomEnabled={false}
                style={{
                    height: "100%",
                    width: "100%"
                }}
            >
            </AnimatedMapView>
        );
    }, [isLoading.current, isActive, colors.COLOR_SCHEME]);

    // If we're still loading the page, return a generic, empty view.
    if (!isActive) {
        return null;
    }

    if (isLoading.current) {
        return (
            <View style={{
                aspectRatio: mapAspectRatio,
                width: "100%",
                borderRadius: style_values.border_radius,
                overflow: "hidden",
                backgroundColor: Color(colors.SECONDARY_TEXT).mix(Color(colors.BACKGROUND), 0.5).hex(),
                justifyContent: "center",
                alignItems: "center"
            }}>
                <StyledText>Retrieving your location...</StyledText>
            </View>
        );
    }

    return (
        <View style={[
            {
                aspectRatio: mapAspectRatio,
                width: "100%",
                borderRadius: style_values.border_radius,
                overflow: "hidden",
                backgroundColor: "red"
            },
            style
        ]}>
            {memoizedMap}
        </View>
    );
};

export default SelfProfileScreen;