import { useColorScheme } from 'react-native';

/**
 * @typedef {object} ColorSpecification
 * @property {"light"|"dark"} COLOR_SCHEME The value returned by `useColorScheme()` to determine if
 *      the user is using light or dark mode.
 * @property {String} FLAT1 The first color of the primary color scheme. Blue.
 * @property {String} FLAT2 The second color of the primary color scheme. Purple.
 * @property {String} FLAT3 The third color of the primary color scheme. Yellow.
 * @property {String} FLAT4 The fourth color of the primary color scheme. Pink.
 * @property {String} BACKGROUND The color to use for background colors of the app. A light color
 *      if the color scheme is "light", otherwise a dark color.
 * @property {String} TEXT The color to use for text. It is made to contrast with the `BACKGROUND`
 *      color.
 * @property {String} INVERTED_TEXT The color to use for text when the background is inverted from
 *      the normal `BACKGROUND` color.
 * @property {String} PRIMARY_TEXT The same as `TEXT`.
 * @property {String} SECONDARY_TEXT The color to use for secondary text that is less important
 *      than other primary text. It is a grayed out version of `TEXT`.
 */

/**
 * Returns a color specification that updates according to whether the user is using light or dark
 * mode.
 * @param {"light"|"dark"} [themeOverride] An optional string indicating the theme to use (either
 *      `"light"` or `"dark"`). If omitted, uses the device's current color scheme.
 * @returns {ColorSpecification}
 */
const useColors = (themeOverride) => {
    const actualColorScheme = useColorScheme();
    const scheme = themeOverride == "light" || themeOverride == "dark" ? themeOverride : actualColorScheme;
    const color = scheme == "dark" ?
        {
            COLOR_SCHEME: "dark",
            FLAT1: "#84C6EC",
            FLAT2: "#A98df6",
            FLAT3: "#FFC534",
            FLAT4: "#EA6787",
            BACKGROUND: "#333A44",
            TEXT: "#FFFFFF",
            INVERTED_TEXT: "#000000",
            SECONDARY_TEXT: "#8A939B"
        } : {
            COLOR_SCHEME: "light",
            FLAT1: "#3498DB",
            FLAT2: "#855EF3",
            FLAT3: "#FFB700",
            FLAT4: "#EA6787",
            BACKGROUND: "#FFFFFF",
            TEXT: "#334455",
            INVERTED_TEXT: "#FFFFFF",
            SECONDARY_TEXT: "#8A939B",
        };
    color.PRIMARY_TEXT = color.TEXT;
    return color;
};

export default useColors;
