import { useSafeAreaInsets } from "react-native-safe-area-context";
import { style_values } from "./styles";

/**
 * A list of numbers that denote the padding/insets to use on each side of a box.
 * @typedef {object} InsetList
 * @property {number} top The inset to use on the top side.
 * @property {number} right The inset to use on the right side.
 * @property {number} bottom The inset to use on the bottom side.
 * @property {number} left The inset to use on the left side.
 */

/**
 * A function that returns safe area insets (similar to `useSafeAreaInsets()`) except that a default
 * inset amount can be provided to use the maximum of the two instead. This allows you to specify
 * your own padding insets, while still ensuring that safeAreas are respected if the desired insets
 * are not enough.
 * @param {InsetList} [{}] insets An object denoting the desired insets to use. If the safe area is
 *      greater than the desired insets for one of the sides, the safe area value will be returned
 *      instead. If omitted, defaults to style_values.screen_padding.
 * @returns {InsetList} The list of insets that uses the max of the safe area insets and the desired
 *      insets.
 */
const usePaddedSafeAreaInsets = (insets) => {
    const safeAreaInsets = useSafeAreaInsets();
    return {
        top:
            Math.max(
                typeof insets?.top == "number" ? insets.top : style_values.screen_padding.vertical,
                safeAreaInsets.top
            ),
        right:
            Math.max(
                typeof insets?.right == "number" ? insets.right : style_values.screen_padding.horizontal,
                safeAreaInsets.right
            ),
        bottom:
            Math.max(
                typeof insets?.bottom == "number" ? insets.bottom : style_values.screen_padding.vertical,
                safeAreaInsets.bottom
            ),
        left:
            Math.max(
                typeof insets?.left == "number" ? insets.left : style_values.screen_padding.horizontal,
                safeAreaInsets.left
            )
    };
}

export default usePaddedSafeAreaInsets;